"""
.. topic:: Levinson module

    .. autosummary::

        LEVINSON

    .. codeauthor:: Thomas Cokelaer, 2011
"""
import numpy


__all__ = ["LEVINSON", "rlevinson"]


def LEVINSON(r, order=None, allow_singularity=False):
    r"""Levinson-Durbin recursion.

    Find the coefficients of a length(r)-1 order autoregressive linear process

    :param r: autocorrelation sequence of length N + 1 (first element being the zero-lag autocorrelation)
    :param order: requested order of the autoregressive coefficients. default is N.
    :param allow_singularity: false by default. Other implementations may be True (e.g., octave)

    :return:
        * the `N+1` autoregressive coefficients :math:`A=(1, a_1...a_N)`
        * the prediction errors
        * the `N` reflections coefficients values

    This algorithm solves the set of complex linear simultaneous equations
    using Levinson algorithm.

    .. math::

        \bold{T}_M \left( \begin{array}{c} 1 \\ \bold{a}_M \end{array} \right) =
        \left( \begin{array}{c} \rho_M \\ \bold{0}_M  \end{array} \right)

    where :math:`\bold{T}_M` is a Hermitian Toeplitz matrix with elements
    :math:`T_0, T_1, \dots ,T_M`.

    .. note:: Solving this equations by Gaussian elimination would
        require :math:`M^3` operations whereas the levinson algorithm
        requires :math:`M^2+M` additions and :math:`M^2+M` multiplications.

    This is equivalent to solve the following symmetric Toeplitz system of
    linear equations

    .. math::

        \left( \begin{array}{cccc}
        r_1 & r_2^* & \dots & r_{n}^*\\
        r_2 & r_1^* & \dots & r_{n-1}^*\\
        \dots & \dots & \dots & \dots\\
        r_n & \dots & r_2 & r_1 \end{array} \right)
        \left( \begin{array}{cccc}
        a_2\\
        a_3 \\
        \dots \\
        a_{N+1}  \end{array} \right)
        =
        \left( \begin{array}{cccc}
        -r_2\\
        -r_3 \\
        \dots \\
        -r_{N+1}  \end{array} \right)

    where :math:`r = (r_1  ... r_{N+1})` is the input autocorrelation vector, and
    :math:`r_i^*` denotes the complex conjugate of :math:`r_i`. The input r is typically
    a vector of autocorrelation coefficients where lag 0 is the first
    element :math:`r_1`.


    .. doctest::

        >>> import numpy; from spectrum import LEVINSON
        >>> T = numpy.array([3., -2+0.5j, .7-1j])
        >>> a, e, k = LEVINSON(T)

    """
    #from numpy import isrealobj
    T0  = numpy.real(r[0])
    T = r[1:]
    M = len(T)

    if order is None:
        M = len(T)
    else:
        assert order <= M, 'order must be less than size of the input data'
        M = order

    realdata = numpy.isrealobj(r)
    if realdata is True:
        A = numpy.zeros(M, dtype=float)
        ref = numpy.zeros(M, dtype=float)
    else:
        A = numpy.zeros(M, dtype=complex)
        ref = numpy.zeros(M, dtype=complex)

    P = T0

    for k in range(0, M):
        save = T[k]
        if k == 0:
            temp = -save / P
        else:
            #save += sum([A[j]*T[k-j-1] for j in range(0,k)])
            for j in range(0, k):
                save = save + A[j] * T[k-j-1]
            temp = -save / P
        if realdata:
            P = P * (1. - temp**2.)
        else:
            P = P * (1. - (temp.real**2+temp.imag**2))
        if P <= 0 and allow_singularity==False:
            raise ValueError("singular matrix")
        A[k] = temp
        ref[k] = temp # save reflection coeff at each step
        if k == 0:
            continue

        khalf = (k+1)//2
        if realdata is True:
            for j in range(0, khalf):
                kj = k-j-1
                save = A[j]
                A[j] = save + temp * A[kj]
                if j != kj:
                    A[kj] += temp*save
        else:
            for j in range(0, khalf):
                kj = k-j-1
                save = A[j]
                A[j] = save + temp * A[kj].conjugate()
                if j != kj:
                    A[kj] = A[kj] + temp * save.conjugate()

    return A, P, ref


def rlevinson(a, efinal):
    """computes the autocorrelation coefficients, R based
    on the prediction polynomial A and the final prediction error Efinal,
    using the stepdown algorithm.

    Works for real or complex data

    :param a:
    :param efinal:

    :return:
        * R, the autocorrelation
        * U  prediction coefficient
        * kr reflection coefficients
        * e errors

    A should be a minimum phase polynomial and A(1) is assumed to be unity.

    :returns: (P+1) by (P+1) upper triangular matrix, U,
        that holds the i'th order prediction polynomials
        Ai, i=1:P, where P is the order of the input
        polynomial, A.



             [ 1  a1(1)*  a2(2)* ..... aP(P)  * ]
             [ 0  1       a2(1)* ..... aP(P-1)* ]
       U  =  [ .................................]
             [ 0  0       0      ..... 1        ]

    from which the i'th order prediction polynomial can be extracted
    using Ai=U(i+1:-1:1,i+1)'. The first row of U contains the
    conjugates of the reflection coefficients, and the K's may be
    extracted using, K=conj(U(1,2:end)).

    .. todo:: remove the conjugate when data is real data, clean up the code
       test and doc.

    """
    a = numpy.array(a)
    realdata = numpy.isrealobj(a)


    assert a[0] == 1, 'First coefficient of the prediction polynomial must be unity'

    p = len(a)

    if p < 2:
        raise ValueError('Polynomial should have at least two coefficients')

    if realdata == True:
        U = numpy.zeros((p, p)) # This matrix will have the prediction
                                # polynomials of orders 1:p
    else:
        U = numpy.zeros((p, p), dtype=complex)
    U[:, p-1] = numpy.conj(a[-1::-1]) # Prediction coefficients of order p

    p = p -1
    e = numpy.zeros(p)

    # First we find the prediction coefficients of smaller orders and form the
    # Matrix U

    # Initialize the step down

    e[-1] = efinal # Prediction error of order p

    # Step down
    for k in range(p-1, 0, -1):
        [a, e[k-1]] = levdown(a, e[k])
        U[:, k] = numpy.concatenate((numpy.conj(a[-1::-1].transpose()) ,
                                      [0]*(p-k) ))




    e0 = e[0]/(1.-abs(a[1]**2)) #% Because a[1]=1 (true polynomial)
    U[0,0] = 1                #% Prediction coefficient of zeroth order
    kr = numpy.conj(U[0,1:])     #% The reflection coefficients
    kr = kr.transpose()                 #% To make it into a column vector

    #   % Once we have the matrix U and the prediction error at various orders, we can
    #  % use this information to find the autocorrelation coefficients.

    R = numpy.zeros(1, dtype=complex)
    #% Initialize recursion
    k = 1
    R0 = e0 # To take care of the zero indexing problem
    R[0] = -numpy.conj(U[0,1])*R0   # R[1]=-a1[1]*R[0]

    # Actual recursion
    for k in range(1,p):
        r = -sum(numpy.conj(U[k-1::-1,k])*R[-1::-1]) - kr[k]*e[k-1]
        R = numpy.insert(R, len(R), r)

    # Include R(0) and make it a column vector. Note the dot transpose

    #R = [R0 R].';
    R = numpy.insert(R, 0, e0)
    return R, U, kr, e


def levdown(anxt, enxt=None):
    """One step backward Levinson recursion

    :param anxt:
    :param enxt:
    :return:
        * acur the P'th order prediction polynomial based on the P+1'th order prediction polynomial, anxt.
        * ecur the the P'th order prediction error  based on the P+1'th order prediction error, enxt.

    ..  * knxt the P+1'th order reflection coefficient.

    """
    #% Some preliminaries first
    #if nargout>=2 & nargin<2
    #    raise ValueError('Insufficient number of input arguments');
    if anxt[0] != 1:
        raise ValueError('At least one of the reflection coefficients is equal to one.')
    anxt = anxt[1:] #  Drop the leading 1, it is not needed
                    #  in the step down

    # Extract the k+1'th reflection coefficient
    knxt = anxt[-1]
    if knxt == 1.0:
        raise ValueError('At least one of the reflection coefficients is equal to one.')

    # A Matrix formulation from Stoica is used to avoid looping
    acur = (anxt[0:-1]-knxt*numpy.conj(anxt[-2::-1]))/(1.-abs(knxt)**2)
    ecur = None
    if enxt is not None:
        ecur = enxt/(1.-numpy.dot(knxt.conj().transpose(),knxt))

    acur = numpy.insert(acur, 0, 1)

    return acur, ecur


def levup(acur, knxt, ecur=None):
    """LEVUP  One step forward Levinson recursion

    :param acur:
    :param knxt:
    :return:
        * anxt the P+1'th order prediction polynomial based on the P'th order prediction polynomial, acur, and the
          P+1'th order reflection coefficient, Knxt.
        * enxt the P+1'th order prediction  prediction error, based on the P'th order prediction error, ecur.


    :References:  P. Stoica R. Moses, Introduction to Spectral Analysis  Prentice Hall, N.J., 1997, Chapter 3.
    """
    if acur[0] != 1:
        raise ValueError('At least one of the reflection coefficients is equal to one.')
    acur = acur[1:] #  Drop the leading 1, it is not needed

    # Matrix formulation from Stoica is used to avoid looping
    anxt = numpy.concatenate((acur, [0])) + knxt * numpy.concatenate((numpy.conj(acur[-1::-1]), [1]))

    enxt = None
    if ecur is not None:
        # matlab version enxt = (1-knxt'.*knxt)*ecur
        enxt = (1. - numpy.dot(numpy.conj(knxt), knxt)) * ecur

    anxt = numpy.insert(anxt, 0, 1)

    return anxt, enxt



"""Correlogram PSD estimates

.. topic:: This module provides Correlograms methods

    .. autosummary::

        CORRELOGRAMPSD
        pcorrelogram

    .. codeauthor:: Thomas Cokelaer 2011

    :References: See [Marple]_
"""
import numpy
from spectrum.correlation import CORRELATION, xcorr
from spectrum.window import Window
from numpy.fft import fft
from spectrum.psd import FourierSpectrum
from spectrum import tools

__all__ = ["CORRELOGRAMPSD", "pcorrelogram"]


def CORRELOGRAMPSD(X, Y=None, lag=-1, window='hamming',
                    norm='unbiased', NFFT=4096, window_params={},
                    correlation_method='xcorr'):
    """PSD estimate using correlogram method.


    :param array X: complex or real data samples X(1) to X(N)
    :param array Y: complex data samples Y(1) to Y(N). If provided, computes
        the cross PSD, otherwise the PSD is returned
    :param int lag: highest lag index to compute. Must be less than N
    :param str window_name: see :mod:`window` for list of valid names
    :param str norm: one of the valid normalisation of :func:`xcorr` (biased, 
        unbiased, coeff, None)
    :param int NFFT: total length of the final data sets (padded with zero 
        if needed; default is 4096)
    :param str correlation_method: either `xcorr` or `CORRELATION`.
        CORRELATION should be removed in the future.

    :return:
        * Array of real (cross) power spectral density estimate values. This is
          a two sided array with negative values following the positive ones
          whatever is the input data (real or complex).

    .. rubric:: Description:

    The exact power spectral density is the Fourier transform of the
    autocorrelation sequence:

    .. math:: P_{xx}(f) = T \sum_{m=-\infty}^{\infty} r_{xx}[m] exp^{-j2\pi fmT}

    The correlogram method of PSD estimation substitutes a finite sequence of
    autocorrelation estimates :math:`\hat{r}_{xx}` in place of :math:`r_{xx}`.
    This estimation can be computed with :func:`xcorr` or :func:`CORRELATION` by
    chosing a proprer lag `L`. The estimated PSD is then

    .. math:: \hat{P}_{xx}(f) = T \sum_{m=-L}^{L} \hat{r}_{xx}[m] exp^{-j2\pi fmT}

    The lag index must be less than the number of data samples `N`. Ideally, it
    should be around `L/10` [Marple]_ so as to avoid greater statistical
    variance associated with higher lags.

    To reduce the leakage of the implicit rectangular window and therefore to
    reduce the bias in the estimate, a tapering window is normally used and lead
    to the so-called Blackman and Tukey correlogram:

    .. math:: \hat{P}_{BT}(f) = T \sum_{m=-L}^{L} w[m] \hat{r}_{xx}[m] exp^{-j2\pi fmT}

    The correlogram for the cross power spectral estimate is

    .. math:: \hat{P}_{xx}(f) = T \sum_{m=-L}^{L} \hat{r}_{xx}[m] exp^{-j2\pi fmT}

    which is computed if :attr:`Y` is not provide. In such case,
    :math:`r_{yx} = r_{xy}` so we compute the correlation only once.

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import CORRELOGRAMPSD, marple_data
        from spectrum.tools import cshift
        from pylab import log10, axis, grid, plot,linspace

        psd = CORRELOGRAMPSD(marple_data, marple_data, lag=15)
        f = linspace(-0.5, 0.5, len(psd))
        psd = cshift(psd, len(psd)/2)
        plot(f, 10*log10(psd/max(psd)))
        axis([-0.5,0.5,-50,0])
        grid(True)

    .. seealso:: :func:`create_window`, :func:`CORRELATION`, :func:`xcorr`,
        :class:`pcorrelogram`.
    """
    N = len(X)
    assert lag<N, 'lag must be < size of input data'
    assert correlation_method in ['CORRELATION', 'xcorr']
    if Y is None:
        Y = numpy.array(X)
        crosscorrelation = False
    else:
        crosscorrelation = True

    if NFFT is None:
        NFFT = N
    psd = numpy.zeros(NFFT, dtype=complex)

    # Window should be centered around zero. Moreover, we want only the
    # positive values. So, we need to use 2*lag + 1 window and keep values on
    # the right side.
    w = Window(2*lag+1, window, **window_params)
    w = w.data[lag+1:]

    # compute the cross correlation
    if correlation_method == 'CORRELATION':
        rxy = CORRELATION (X, Y, maxlags=lag, norm=norm)
    elif correlation_method == 'xcorr':
        rxy, _l = xcorr (X, Y, maxlags=lag, norm=norm)
        rxy = rxy[lag:]

    # keep track of the first elt.
    psd[0] = rxy[0]

    # create the first part of the PSD
    psd[1:lag+1] = rxy[1:] * w

    # create the second part.
    # First, we need to compute the auto or cross correlation ryx
    if crosscorrelation is True:
        # compute the cross correlation
        if correlation_method == 'CORRELATION':
            ryx = CORRELATION(Y, X, maxlags=lag, norm=norm)
        elif correlation_method == 'xcorr':
            ryx, _l = xcorr(Y, X, maxlags=lag, norm=norm)
            ryx = ryx[lag:]
        #print len(ryx), len(psd[-1:NPSD-lag-1:-1])

        psd[-1:NFFT-lag-1:-1] = ryx[1:].conjugate() * w
    else: #autocorrelation no additional correlation call required
        psd[-1:NFFT-lag-1:-1] = rxy[1:].conjugate() * w

    psd = numpy.real(fft(psd))

    return psd


class pcorrelogram(FourierSpectrum):
    """The Correlogram class provides an interface to :func:`CORRELOGRAMPSD`.

    It returns an object that inherits from :class:`FourierSpectrum` and
    therefore ease the manipulation of PSDs.

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import pcorrelogram, data_cosine
        p = pcorrelogram(data_cosine(N=1024), lag=15)
        p.plot()
        p.plot(sides='twosided')

    """
    def __init__(self, data, sampling=1., lag=-1,
                 window='hamming', NFFT=None, scale_by_freq=True,
                 detrend=None):
        """**Correlogram Constructor**

        :param array data: input data (list or numpy.array)
        :param float sampling: sampling frequency of the input :attr:`data`.
        :param int lag:
        :param str window: a tapering window. See :class:`~spectrum.window.Window`.
        :param int NFFT: total length of the final data sets (padded with 
            zero if needed; default is 4096)
        :param bool scale_by_freq:
        :param str detrend:
        """
        super(pcorrelogram, self).__init__(data,
                                          window=window,
                                          sampling=sampling,
                                          NFFT=NFFT,
                                          scale_by_freq=scale_by_freq,
                                          lag=lag,
                                          detrend=detrend)

    def __call__(self):
        psd = CORRELOGRAMPSD(self.data, self.data_y,
                             lag=self.lag,
                             window=self.window,
                             NFFT=self.NFFT,
                             #   scale_by_freq=self.scale_by_freq,
                             )
        if self.datatype == 'real':
            #FIXME. do we want to use same syntax/code as in burg/pminvar/pcovar
            # to handle odd data ?
            self.psd = tools.twosided_2_onesided(psd)
        else:
            self.psd = psd
        self.scale()
        return self

    def _str_title(self):
        return "Correlogram PSD estimate\n"

    def __str__(self):
        return super(pcorrelogram, self).__str__()




"""Yule Walker method to estimate AR values.

.. topic:: Estimation of AR values using Yule-Walker method

    .. autosummary::

        aryule
        pyule

    .. codeauthor:: Thomas Cokelaer 2011

"""

from .correlation import CORRELATION
from .levinson import LEVINSON
from .psd import ParametricSpectrum
from spectrum import tools
import numpy as np

__all__ = ['aryule', 'pyule']


def aryule(X, order, norm='biased', allow_singularity=True):
    r"""Compute AR coefficients using Yule-Walker method

    :param X: Array of complex data values, X(1) to X(N)
    :param int order: Order of autoregressive process to be fitted (integer)
    :param str norm: Use a biased or unbiased correlation.
    :param bool allow_singularity:

    :return:
        * AR coefficients (complex)
        * variance of white noise (Real)
        * reflection coefficients for use in lattice filter

    .. rubric:: Description:

    The Yule-Walker method returns the polynomial A corresponding to the
    AR parametric signal model estimate of vector X using the Yule-Walker
    (autocorrelation) method. The autocorrelation may be computed using a
    **biased** or **unbiased** estimation. In practice, the biased estimate of
    the autocorrelation is used for the unknown true autocorrelation. Indeed,
    an unbiased estimate may result in nonpositive-definite autocorrelation
    matrix.
    So, a biased estimate leads to a stable AR filter.
    The following matrix form represents the Yule-Walker equations. The are
    solved by means of the Levinson-Durbin recursion:

     .. math::

        \left( \begin{array}{cccc}
        r(1) & r(2)^* & \dots & r(n)^*\\
        r(2) & r(1)^* & \dots & r(n-1)^*\\
        \dots & \dots & \dots & \dots\\
        r(n) & \dots & r(2) & r(1) \end{array} \right)
        \left( \begin{array}{cccc}
        a(2)\\
        a(3) \\
        \dots \\
        a(n+1)  \end{array} \right)
        =
        \left( \begin{array}{cccc}
        -r(2)\\
        -r(3) \\
        \dots \\
        -r(n+1)  \end{array} \right)

    The outputs consists of the AR coefficients, the estimated variance of the
    white noise process, and the reflection coefficients. These outputs can be
    used to estimate the optimal order by using :mod:`~spectrum.criteria`.

    .. rubric:: Examples:

    From a known AR process or order 4, we estimate those AR parameters using
    the aryule function.

    .. doctest::

        >>> from scipy.signal import lfilter
        >>> from spectrum import *
        >>> from numpy.random import randn
        >>> A  =[1, -2.7607, 3.8106, -2.6535, 0.9238]
        >>> noise = randn(1, 1024)
        >>> y = lfilter([1], A, noise);
        >>> #filter a white noise input to create AR(4) process
        >>> [ar, var, reflec] = aryule(y[0], 4)
        >>> # ar should contains values similar to A

    The PSD estimate of a data samples is computed and plotted as follows:

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import *
        from pylab import *

        ar, P, k = aryule(marple_data, 15, norm='biased')
        psd = arma2psd(ar)
        plot(linspace(-0.5, 0.5, 4096), 10 * log10(psd/max(psd)))
        axis([-0.5, 0.5, -60, 0])

    .. note:: The outputs have been double checked against (1) octave outputs
        (octave has norm='biased' by default) and (2) Marple test code.

    .. seealso:: This function uses :func:`~spectrum.levinson.LEVINSON` and
        :func:`~spectrum.correlation.CORRELATION`. See the :mod:`~spectrum.criteria`
        module for criteria to automatically select the AR order.

    :References: [Marple]_

    """
    assert norm in ['biased', 'unbiased']
    r = CORRELATION(X, maxlags=order, norm=norm)
    A, P, k = LEVINSON(r, allow_singularity=allow_singularity)
    return A, P, k


class pyule(ParametricSpectrum):
    """Class to create PSD based on the Yule Walker method

    See :func:`aryule` for description.

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import *
        p = pyule(marple_data, 15, NFFT=4096)
        p.plot(sides='centerdc')

    """
    def __init__(self, data, order, norm='biased', NFFT=None, sampling=1.,
                 scale_by_freq=True):
        """**Constructor**

        For a detailled description of the parameters, see :func:`aryule`.

        :param array data: input data (list or numpy.array)
        :param int order:
        :param int NFFT: total length of the final data sets (padded with
            zero if needed; default is 4096)
        :param float sampling: sampling frequency of the input :attr:`data`
        :param str norm: don't change if you do not know

        """
        super(pyule, self).__init__(data, ar_order=order,  NFFT=NFFT,
                                    scale_by_freq=scale_by_freq,
                                    sampling=sampling)
        self.sampling = sampling
        self._norm_aryule = norm

    def __call__(self):
        from . import arma
        ar, rho, k = aryule(self.data, self.ar_order, norm=self._norm_aryule)
        psd = arma.arma2psd(ar, NFFT=self.NFFT, rho=rho, T=self.sampling)
        # save the AR and reflection coefficients.
        self.ar = ar
        self.reflection = k

        # save the PSD
        if self.datatype == 'real':
            # see doc/concepts.rst for details
            if self.NFFT % 2 == 0:
                newpsd  = psd[0:int(self.NFFT/2 + 1)] * 2
            else:
                newpsd  = psd[0:int((self.NFFT+1) / 2)] * 2
            self.psd = newpsd
        else:
            self.psd = psd
        if self.scale_by_freq is True:
            self.scale()
        return self

    def _str_title(self):
        return "PYule PSD estimate\n"

    def __str__(self):
        return super(pyule, self).__str__()





__all__ = ['readwav']


def readwav(filename):
    """Read a WAV file and returns the data and sample rate

    ::

        from spectrum.io import readwav
        readwav()

    """
    from scipy.io.wavfile import read as readwav
    samplerate, signal = readwav(filename)
    return signal, samplerate



from spectrum import Periodogram, pmtm
import numpy as np


__all__ = ["Spectrogram"]


class Spectrogram(object):
    """Simple example of spectrogram

    .. plot::

        from spectrum import Spectrogram, dolphin_filename, readwav
        data, samplerate = readwav(dolphin_filename)

        p = Spectrogram(data, ws=128, W=4096, sampling=samplerate)
        p.periodogram()
        p.plot()

    .. warning:: this is a prototype and need careful checking about x/y axis


    """
    def __init__(self, signal, ws=128, W=4096, sampling=1, channel=1):
        if len(signal.shape) == 1:
            self.signal = signal
        else:
            self.signal = signal[:,channel-1]
        self.W = W
        self.ws = ws
        self._start_y = 10
        self.sampling = sampling
        self.duration = len(self.signal) / float(self.sampling)

    def plot(self, filename=None, vmin=None, vmax=None, cmap='jet_r'):
        import pylab
        pylab.clf()
        pylab.imshow(-np.log10(self.results[self._start_y:,:]), 
            origin="lower",
            aspect="auto", cmap=cmap, vmin=vmin, vmax=vmax)
        pylab.colorbar()

        # Fix xticks
        XMAX = float(self.results.shape[1])  # The max integer on xaxis
        xpos = list(range(0, int(XMAX), int(XMAX/5)))
        xx = [int(this*100)/100 for this in np.array(xpos) / XMAX * self.duration]
        pylab.xticks(xpos, xx, fontsize=16)

        # Fix yticks
        YMAX = float(self.results.shape[0])  # The max integer on xaxis
        ypos = list(range(0, int(YMAX), int(YMAX/5)))
        yy = [int(this) for this in np.array(ypos) / YMAX * self.sampling]
        pylab.yticks(ypos, yy, fontsize=16)

        #pylab.yticks([1000,2000,3000,4000], [5500,11000,16500,22000], fontsize=16)
        #pylab.title("%s echoes" %  filename.replace(".png", ""), fontsize=25)
        pylab.xlabel("Time (seconds)", fontsize=25)
        pylab.ylabel("Frequence (Hz)", fontsize=25)
        pylab.tight_layout()
        if filename:
            pylab.savefig(filename)

    def periodogram(self):
        W = self.W
        ws = self.ws
        N = int(len(self.signal)/ws)
        self.results = np.zeros((W*2+1, N-8))
        print("Duration: %s" % self.duration)
        print("W: %s" % W)
        print("ws: %s" % ws)
        print("Computing %s TFs" % N)
        for i in range(N-8):
            data = self.signal[i*ws:i*ws+W]
            p = Periodogram(data, sampling=self.sampling, NFFT=W*4)
            p()
            self.results[:,i] = p.psd
        print("done")

    def pmtm(self):
        W = self.W
        ws = self.ws
        N = int(len(self.signal) / ws)
        self.results = np.zeros((W+1, N-8))
        for i in range(N-8):
            data = self.signal[i*ws:i*ws+W]
            a = pmtm(data, 4, NFFT=W*4, show=False)
            Sk = np.mean(abs(a[0].transpose())**2 * a[1], axis=1)
            self.results[:, i] = Sk[0:self.W+1]
            print(i, N)
        print("done")





"""Criteria for parametric methods.

.. topic:: This module provides criteria to automatically select order in
    parametric PSD estimate or pseudo spectrum estimates (e.g, music).

    Some criteria such as the AIC criterion helps to chose the order of PSD
    models such as the ARMA model. Nevertheless, it is difficult to estimate
    correctly the order of an ARMA model even by using these criteria. The
    reason being that even the Akaike criteria (AIC) does not provide the
    proper order with a probability of 1 with infinite samples.

    The order choice is related to an expertise of the signal. There is no
    exact criteria. However, they may provide useful information.

    AIC, AICc, KIC and AKICc are based on information theory.  They  attempt
    to balance the complexity (or length) of the model against how well the
    model fits the data.  AIC and KIC are biased estimates of the asymmetric
    and the symmetric Kullback-Leibler divergence respectively.  AICc and
    AKICc attempt to correct the bias.

    There are also criteria related to eigen analysis, which takes as input
    the eigen values of any PSD estimate method.

.. rubric:: Example

.. plot::
    :width: 80%
    :include-source:

    from spectrum import aryule, AIC, marple_data
    from pylab import plot, arange
    order = arange(1, 25)
    rho = [aryule(marple_data, i, norm='biased')[1] for i in order]
    plot(order, AIC(len(marple_data), rho, order), label='AIC')



:References: bd-Krim Seghouane and Maiza Bekara
   "A small sample model selection criterion based on Kullback's symmetric
   divergence", IEEE Transactions on Signal Processing,
   Vol. 52(12), pp 3314-3323, Dec. 2004

"""


class Criteria(object):
    """Criteria class for an automatic selection of ARMA order.

    Available criteria are

    ======= =====================
    ======= =====================
    AIC     see :func:`AIC`
    AICc    see :func:`AICc`
    KIC     see :func:`KIC`
    AKICc   see :func:`AKICc`
    FPE     see :func:`FPE`
    MDL     see :func:`MDL`
    CAT     see :func:`_CAT`
    ======= =====================


    """
    valid_criteria_names =  ['AIC', 'AICc', 'KIC', 'FPE', 'AKICc', 'MDL']
    error_incorrect_name = 'Invalid name provided. Correct names are %s ' \
        % valid_criteria_names
    error_no_criteria_found = 'No names match the valid criteria names (%s)' \
        % valid_criteria_names
    def __init__(self, name, N):
        """Create a criteria object

        :param name: a string or list of strings containing valid criteria
            method's name
        :param int N: size of the data sample.

        """
        #valid attributes
        self.__name = None
        self.name = name
        self.__N = N
        self.__rho = 0
        self.__k = None
        self.__old_data = None
        self.__data = None
        self.__norm = True

    def _getName(self):
        return self.__name
    def _setName(self, name):
        assert isinstance(name, str), 'name must be a string'
        if name in self.valid_criteria_names:
            self.__name = name
        else:
            raise ValueError(self.error_no_criteria_found)
    name = property(fget=_getName, fset=_setName, doc="Getter/Setter for the criteria name")

    def _getData(self):
        return self.__data
    def _setData(self, data):
        # save the data value in old_data is there is something to save
        if self.data is None:
            self.__data = data
            self.__old_data = 2.*data
        else:
            self.__old_data = self.data
            self.__data = data
    data = property(fget=_getData, fset=_setData, doc="Getter/Setter for the criteria output")

    def _getOldData(self):
        return self.__old_data
    old_data = property(fget=_getOldData, doc="Getter/Setter for the previous value")

    def _getK(self):
        return self.__k
    k = property(fget=_getK, doc="Getter for k the order of evaluation")

    def _getN(self):
        return self.__N
    def _setN(self, N):
        assert N > 0, 'N must be positive'
        self.__N = N
    N = property(fget=_getN, fset=_setN, doc="Getter/Setter for N")

    def _getRho(self):
        return self.__rho
    def _setRho(self, rho):
        self.__rho = rho
    rho = property(fget=_getRho, fset=_setRho, doc="Getter/Setter for rho")

    def __call__(self, rho=None, k=None, N=None, norm=True):
        """Call the criteria function corresponding to :attr:`name`."""
        self.__norm = norm
        if N is not None:
            self.N = N

        # we update rho only if it is needed (input different from self.rho)
        # if such case, we also update k
        if rho is not None:
            self.rho = rho
        if k is not None:
            self.__k = k
        self.__norm = norm
        #used to check if the criteria is reached or not

        f = eval(self.name)
        self.data = f(self.N, self.rho, self.k)
        # compare the new data with the previous one and return
        # False if the new value is larger so as to stop the iteration
        if self.old_data is not None and self.data is not None:
            if self.data > self.old_data:
                return False
            else:
                return True
        return True


def AIC(N, rho, k):
    r"""Akaike Information Criterion

    :param rho: rho at order k
    :param N: sample size
    :param k: AR order.

    If k is the AR order and N the size of the sample, then Akaike criterion is

    .. math:: AIC(k) = \log(\rho_k) + 2\frac{k+1}{N}

    ::

        AIC(64, [0.5,0.3,0.2], [1,2,3])

    :validation: double checked versus octave.
    """
    from numpy import log, array
    #k+1 #todo check convention. agrees with octave

    res = N * log(array(rho)) + 2.* (array(k)+1)
    return res


def AICc(N, rho, k, norm=True):
    r"""corrected Akaike information criterion

    .. math:: AICc(k) = log(\rho_k) + 2 \frac{k+1}{N-k-2}


    :validation: double checked versus octave.
    """
    from numpy import log, array
    p = k  #todo check convention. agrees with octave
    res = log(rho) + 2. * (p+1) / (N-p-2)
    return res


def KIC(N, rho, k):
    r"""Kullback information criterion

    .. math:: KIC(k) = log(\rho_k) + 3 \frac{k+1}{N}

    :validation: double checked versus octave.
    """
    from numpy import log, array
    res = log(rho) + 3. * (k+1.) /float(N)
    return res


def AKICc(N, rho, k):
    r"""approximate corrected Kullback information

    .. math:: AKICc(k) = log(rho_k) + \frac{p}{N*(N-k)} + (3-\frac{k+2}{N})*\frac{k+1}{N-k-2}

    """
    from numpy import log, array
    p = k
    res = log(rho) + p/N/(N-p) + (3.-(p+2.)/N) * (p+1.) / (N-p-2.)
    return res


def FPE(N,rho, k=None):
    r"""Final prediction error criterion

    .. math:: FPE(k) = \frac{N + k + 1}{N - k - 1} \rho_k

    :validation: double checked versus octave.

    """
    #k #todo check convention. agrees with octave
    fpe = rho * (N + k + 1.) / (N- k -1)
    return fpe


def MDL(N, rho, k):
    r"""Minimum Description Length

    .. math:: MDL(k) = N log \rho_k + p \log N

    :validation: results
    """
    from numpy import log
    #p = arange(1, len(rho)+1)
    mdl = N* log(rho) + k * log(N)
    return mdl


def CAT(N, rho, k):
    r"""Criterion Autoregressive Transfer Function :

    .. math::  CAT(k) = \frac{1}{N} \sum_{i=1}^k \frac{1}{\rho_i} - \frac{\rho_i}{\rho_k}

    .. todo:: validation
    """
    from numpy import zeros, arange
    cat = zeros(len(rho))
    for p in arange(1, len(rho)+1):
        rho_p = float(N)/(N-p)*rho[p-1]
        s = 0
        for j in range(1, p+1):
            rho_j = float(N)/(N-j)*rho[j-1]
            s = s + 1./rho_j
        #print(s, s/float(N), 1./rho_p)
        cat[p-1] = s/float(N) - 1./rho_p
    return cat


def aic_eigen(s, N):
    r"""AIC order-selection using eigen values

    :param s: a list of `p` sorted eigen values
    :param N: the size of the input data. To be defined precisely.

    :return:
        * an array containing the AIC values

    Given :math:`n` sorted eigen values :math:`\lambda_i` with
    :math:`0 <= i < n`, the proposed criterion from Wax and Kailath (1985)
    is:

    .. math:: AIC(k) = -2(n-k)N \ln \frac{g(k)}{a(k)} + 2k(2n-k)

    where the arithmetic sum :math:`a(k)` is:

    .. math:: a(k) = \sum_{i=k+1}^{n}\lambda_i

    and the geometric sum :math:`g(k)` is:

    .. math:: g(k) = \prod_{i=k+1}^{n} \lambda_i^{-(n-k)}

    The number of relevant sinusoids in the signal subspace is determined by
    selecting the minimum of `AIC`.

    .. seealso:: :func:`~spectrum.eigenfreq.eigen`
    .. todo:: define precisely the input parameter N. Should be the input
       data length but when using correlation matrix (SVD), I suspect it
       should be the length of the correlation matrix rather than the
       original data.

    :References:
        * [Marple]_ Chap 13,
        * [Wax]_
    """
    import numpy as np

    kaic = []
    n = len(s)
    for k in range(0, n-1):
        ak = 1./(n-k) * np.sum(s[k+1:])
        gk = np.prod(s[k+1:]**(1./(n-k)))
        kaic.append( -2.*(n-k)*N * np.log(gk/ak) + 2.*k*(2.*n-k))

    return kaic


def mdl_eigen(s, N):
    r"""MDL order-selection using eigen values

    :param s: a list of `p` sorted eigen values
    :param N: the size of the input data. To be defined precisely.

    :return:
        * an array containing the AIC values

    .. math:: MDL(k) = (n-k)N \ln \frac{g(k)}{a(k)} + 0.5k(2n-k) log(N)

    .. seealso:: :func:`aic_eigen` for details

    :References:
        * [Marple]_ Chap 13,
        * [Wax]_
    """
    import numpy as np
    kmdl = []
    n = len(s)
    for k in range(0, n-1):
        ak = 1./(n-k) * np.sum(s[k+1:])
        gk = np.prod(s[k+1:]**(1./(n-k)))
        kmdl.append( -(n-k)*N * np.log(gk/ak) + 0.5*k*(2.*n-k)*np.log(N))
    return kmdl




"""BURG method of AR model estimate

.. topic:: This module provides BURG method and BURG PSD estimate

    .. autosummary::

        arburg
        pburg

    .. codeauthor:: Thomas Cokelaer 2011

"""
#TODO: convert arburg into arburg2 to get a nicer and faster algorithm.
import logging

import numpy as np
from spectrum.psd import ParametricSpectrum

__all__ = ["arburg", 'pburg']


def _arburg2(X, order):
    """This version is 10 times faster than arburg, but the output rho is not correct.


    returns [1 a0,a1, an-1]

    """
    x = np.array(X)
    N = len(x)

    if order <= 0.:
        raise ValueError("order must be > 0")

    # Initialisation
    # ------ rho, den
    rho = sum(abs(x)**2.) / N  # Eq 8.21 [Marple]_
    den = rho * 2. * N

    # ------ backward and forward errors
    ef = np.zeros(N, dtype=complex)
    eb = np.zeros(N, dtype=complex)
    for j in range(0, N):  #eq 8.11
        ef[j] = x[j]
        eb[j] = x[j]

    # AR order to be stored
    a = np.zeros(1, dtype=complex)
    a[0] = 1
    # ---- rflection coeff to be stored
    ref = np.zeros(order, dtype=complex)

    temp = 1.
    E = np.zeros(order+1)
    E[0] = rho

    for m in range(0, order):
        #print m
        # Calculate the next order reflection (parcor) coefficient
        efp = ef[1:]
        ebp = eb[0:-1]
        #print efp, ebp
        num = -2.* np.dot(ebp.conj().transpose(),  efp)
        den = np.dot(efp.conj().transpose(),  efp)
        den += np.dot(ebp,  ebp.conj().transpose())
        ref[m] = num / den

        # Update the forward and backward prediction errors
        ef = efp + ref[m] * ebp
        eb = ebp + ref[m].conj().transpose() * efp

        # Update the AR coeff.
        a.resize(len(a)+1, refcheck=False)
        a = a + ref[m] * np.flipud(a).conjugate()

        # Update the prediction error
        E[m+1] = (1 - ref[m].conj().transpose()*ref[m]) * E[m]
        #print 'REF', ref, num, den
    return a, E[-1], ref


class pburg(ParametricSpectrum):
    """Class to create PSD based on Burg algorithm

    See :func:`arburg` for description.

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import *
        p = pburg(marple_data, 15, NFFT=4096)
        p.plot(sides='centerdc')

    Another example based on a real data set is shown here below. Note here
    that we set the scale_by_freq value to False and True. False should give
    results equivalent to octave or matlab convention while setting to True
    implies that the data is multiplied by  :math:`2\pi df` where 
    :math:`df = sampling / N`.

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import data_two_freqs, pburg
        p = pburg(data_two_freqs(), 7, NFFT=4096)
        p.plot()
        p = pburg(data_two_freqs(), 7, NFFT=4096, scale_by_freq=True)
        p.plot()
        from pylab import legend
        legend(["un-scaled", "scaled by 2 pi df"])

    """
    def __init__(self, data, order, criteria=None, NFFT=None, sampling=1.,
        scale_by_freq=False):
        """**Constructor**

        For a detailled description of the parameters, see :func:`burg`.

        :param array data:     input data (list or np.array)
        :param int order:
        :param str criteria:
        :param int NFFT: total length of the final data sets (padded with zero if 
            needed; default is 4096)

        :param float sampling: sampling frequency of the input :attr:`data`.

        """
        super(pburg, self).__init__(data, ar_order=order,
                sampling=sampling, NFFT=NFFT,
                scale_by_freq=scale_by_freq)
        self.criteria = criteria

    def __call__(self):
        from .arma import arma2psd
        ar, rho, ref = arburg(self.data, self.ar_order, self.criteria)
        self.ar = ar
        self.rho = rho
        self.reflection = ref
        psd = arma2psd(A=self.ar, B=self.ma, rho=self.rho,
                      T=self.sampling, NFFT=self.NFFT)

        if self.datatype == 'real':
            #  see doc/concepts.rst for details
            if self.NFFT % 2 == 0:
                newpsd  = psd[0:int(self.NFFT/2 +1)] * 2
            else:
                newpsd  = psd[0:int((self.NFFT+1)/2)] * 2
            self.psd = newpsd
        else:
            self.psd = psd
        self.scale()
        return self

    def _str_title(self):
        return "Periodogram PSD estimate\n"

    def __str__(self):
        return super(pburg, self).__str__()


def arburg(X, order, criteria=None):
    r"""Estimate the complex autoregressive parameters by the Burg algorithm.

    .. math:: x(n) = \sqrt{(v}) e(n) + \sum_{k=1}^{P+1} a(k) x(n-k)

    :param x:  Array of complex data samples (length N)
    :param order: Order of autoregressive process (0<order<N)
    :param criteria: select a criteria to automatically select the order

    :return:
        * A Array of complex autoregressive parameters A(1) to A(order). First
          value (unity) is not included !!
        * P Real variable representing driving noise variance (mean square
          of residual noise) from the whitening operation of the Burg
          filter.
        * reflection coefficients defining the filter of the model.

    .. plot::
        :width: 80%
        :include-source:

        from pylab import plot, log10, linspace, axis
        from spectrum import *

        AR, P, k = arburg(marple_data, 15)
        PSD = arma2psd(AR, sides='centerdc')
        plot(linspace(-0.5, 0.5, len(PSD)), 10*log10(PSD/max(PSD)))
        axis([-0.5,0.5,-60,0])

    .. note::
        1. no detrend. Should remove the mean trend to get PSD. Be careful if
           presence of large mean.
        2. If you don't know what the order value should be, choose the
           criterion='AKICc', which has the least bias and best
           resolution of model-selection criteria.

    .. note:: real and complex results double-checked versus octave using
        complex 64 samples stored in marple_data. It does not agree with Marple
        fortran routine but this is due to the simplex precision of complex
        data in fortran.

    :reference: [Marple]_ [octave]_
    """
    if order <= 0.:
        raise ValueError("order must be > 0")

    if order > len(X):
        raise ValueError("order must be less than length input - 2")

    x = np.array(X)
    N = len(x)

    # Initialisation
    # ------ rho, den
    rho = sum(abs(x)**2.) / float(N)  # Eq 8.21 [Marple]_
    den = rho * 2. * N

    # ---- criteria
    if criteria:
        from spectrum import Criteria
        crit = Criteria(name=criteria, N=N)
        crit.data = rho
        logging.debug('Step {}. old criteria={} new one={}.  rho={}'.format(
                0, crit.old_data, crit.data, rho))

    #p =0
    a = np.zeros(0, dtype=complex)
    ref = np.zeros(0, dtype=complex)
    ef = x.astype(complex)
    eb = x.astype(complex)
    temp = 1.
    #   Main recursion

    for k in range(0, order):

        # calculate the next order reflection coefficient Eq 8.14 Marple
        num = sum([ef[j]*eb[j-1].conjugate() for j in range(k+1, N)])
        den = temp * den - abs(ef[k])**2 - abs(eb[N-1])**2
        kp = -2. * num / den #eq 8.14

        temp = 1. - abs(kp)**2.
        new_rho = temp * rho

        if criteria:
            logging.debug('Step {}. old criteria={} new one={}. rho={}'.format(
                k+1, crit.old_data, crit.data, new_rho))
            #k+1 because order goes from 1 to P whereas k starts at 0.
            status = crit(rho=temp*rho, k=k+1)
            if status is False:
                logging.debug('Stop criteria reached %s %s ' % (crit.data, crit.old_data))
                break
        # this should be after the criteria
        rho = new_rho
        if rho <= 0:
            raise ValueError("Found a negative value (expected positive strictly) %s. Decrease the order" % rho)

        a.resize(a.size+1,refcheck=False)
        a[k] = kp
        if k == 0:
            for j in range(N-1, k, -1):
                save2 = ef[j]
                ef[j] = save2 + kp * eb[j-1]          # Eq. (8.7)
                eb[j] = eb[j-1] + kp.conjugate() *  save2

        else:
            # update the AR coeff
            khalf = (k+1)//2  # FIXME here khalf must be an integer
            for j in range(0, khalf):
                ap = a[j] # previous value
                a[j] = ap + kp * a[k-j-1].conjugate()      # Eq. (8.2)
                if j != k-j-1:
                    a[k-j-1] = a[k-j-1] + kp * ap.conjugate()    # Eq. (8.2)

            # update the prediction error
            for j in range(N-1, k, -1):
                save2 = ef[j]
                ef[j] = save2 + kp * eb[j-1]          # Eq. (8.7)
                eb[j] = eb[j-1] + kp.conjugate() *  save2

        # save the reflection coefficient
        ref.resize(ref.size+1, refcheck=False)
        ref[k] = kp

    return a, rho, ref



"""This module provides the Base class for PSDs"""
import logging

import numpy

from spectrum.tools import nextpow2
from spectrum import errors
from spectrum.window import window_names
from spectrum import tools as stools


__all__ = ["Spectrum", "FourierSpectrum", "ParametricSpectrum", "Range"]


class Range(object):
    """A class to ease the creation of frequency ranges.

    Given the length :attr:`N` of a data sample and a sampling frequency
    :attr:`sampling`, this class provides methods to generate frequency
    ranges

        * :meth:`centerdc`: frequency range from -sampling/2 up to sampling/2 (excluded),
        * :meth:`twosided`: frequency range from 0 up to sampling (excluded),
        * :meth:`onesided`: frequency range from 0 up to sampling (included or
          excluded depending on evenness of the data). If NFFT is even, PSD 
          has length NFFT/2 + 1 over the interval [0,pi]. If NFFT is odd, 
          the length of PSD is (NFFT+1)/2 and the interval is [0, pi)


    Each method has a generator version:

    .. doctest::
        :options: +SKIP

        >>> r = Range(10, sampling=1)
        >>> list(r.onesided_gen())
        [0.0, 0.1, 0.2, 0.3, 0.4, 0.5]
        >>> r.onesided()
        [0.0, 0.1, 0.2, 0.3, 0.4, 0.5]

    The frequency range length is :math:`N/2+1` for the *onesided* case
    :math:`((N+1)/2` if :math:`N` is odd), and :math:`N` for the *twosided* and
    *centerdc* cases:

    .. doctest::
        :options: +SKIP

        >>> r.twosided()
        [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
        >>> len(r.twosided())
        10
        >>> len(r.centerdc())
        10
        >>> len(r.onesided())
        5
    """

    def __init__(self, N, sampling=1.):
        """.. rubric:: **Constructor**

        :param int N: the data length
        :param float sampling: sampling frequency of the input :attr:`data`.


        .. rubric:: Attributes:

        From the input parameters, read/write attributes are set:

            * :attr:`N`, the data length,
            * :attr:`sampling`, the sampling frequency.

        Additionally, the following read-only attribute is available:

            * :attr:`df`, the frequency step computed from :attr:`N` and
              :attr:`sampling`.

        """
        self.__N = N
        self.__sampling = sampling
        self.__df = None
        self._setN(N)
        self._setsampling(sampling)

    def _getdf(self):
        return self.__df
    df = property(fget=_getdf, doc="""Getter to access the frequency step,
        computed from :attr:`N` and :attr:`sampling`.""")

    def _getN(self):
        return self.__N
    def _setN(self,N):
        self.__N = N
        self.__df = self.__sampling/float(self.__N)
    N = property(fget=_getN, fset=_setN, doc="""Getter/Setter of the data length.
        If changed, :attr:`df` is updated.""")

    def _getsampling(self):
        return self.__sampling
    def _setsampling(self, sampling):
        self.__sampling = sampling
        self.__df = self.__sampling/float(self.__N)
    sampling = property(fget=_getsampling, fset=_setsampling,
        doc="""Getter/Setter of the sampling frequency. If changed, :attr:`df` is updated.""")

    def centerdc_gen(self):
        """Return the centered frequency range as a generator.

        ::

            >>> print(list(Range(8).centerdc_gen()))
            [-0.5, -0.375, -0.25, -0.125, 0.0, 0.125, 0.25, 0.375]

        """
        for a in range(0, self.N):
            yield (a-self.N/2) * self.df

    def twosided_gen(self):
        """Returns the twosided frequency range as a generator

        ::

            >>> print(list(Range(8).centerdc_gen()))
            [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875]

        """
        for a in range(0, self.N):
            yield a * self.df

    def onesided_gen(self):
        """Return the one-sided frequency range as a generator.

        If :attr:`N` is even, the length is N/2 + 1.
        If :attr:`N` is odd, the length is (N+1)/2.

        ::

            >>> print(list(Range(8).onesided()))
            [0.0, 0.125, 0.25, 0.375, 0.5]
            >>> print(list(Range(9).onesided()))
            [0.0, 0.1111, 0.2222, 0.3333, 0.4444]

        """
        if self.N % 2 == 0:
            for n in range(0, self.N//2 + 1):
                yield n * self.df
        else:
            for n in range(0, (self.N+1)//2):
                yield n * self.df

    def onesided(self):
        """Return the one-sided frequency range as a list (see
        :meth:`onesided_gen` for details).
        """
        return list(self.onesided_gen())

    def twosided(self):
        """Return the two-sided frequency range as a list (see
        :meth:`twosided_gen` for details)."""
        return list(self.twosided_gen())

    def centerdc(self):
        """Return the two-sided frequency range as a list (see
        :meth:`centerdc_gen` for details)."""
        return list(self.centerdc_gen())

    def __str__(self):
        msg = 'Range object information\n'
        msg += '    N=%s\n' % self.__N
        msg += '    sampling=%s\n' % self.__sampling
        msg += '    df=%s\n' % self.__df
        return msg


class Spectrum(object):
    """Base class for all Spectrum classes

    All PSD classes should inherits from this class to store common attributes
    such as the input data or sampling frequency. An instance is created
    as follows::

        >>> p = Spectrum(data, sampling=1024)
        >>> p.data
        >>> p.sampling

    The input parameters are:

    :param array data: input data (list or numpy.array)
    :param array data_y: input data required to perform cross-PSD only.
    :param float sampling: sampling frequency of the input :attr:`data`
    :param str detrend: detrend method ([None,'mean']) to apply on the input data before
        computing the PSD. See :attr:`detrend`.
    :param bool scale_by_freq: divide the final PSD by :math:`2*\pi/df`
    :param int NFFT: total length of the final data sets (padded with 
        zero if needed; default is 4096)

    The input parameters are available as attributes. Additional
    attributes such as :attr:`N` (the data length), :attr:`df` (the frequency
    step are set (see constructor documentation for a complete list).

    .. warning:: :class:`Spectrum` does not compute the PSD estimate.

    You can populate manually the :attr:`psd` attribute but you should
    respect the following convention:

        * if the input data is real, the PSD is assumed to be one-sided (odd
          length)
        * if the input data is complex, the PSD is assumed to be two-sided
          (even length).

    When :attr:`psd` is set, :attr:`sides` is reset to its default value,
    :attr:`NFFT` and :attr:`df` are updated.

    Spectrum instances have plotting utilities like :meth:`plot`
    that take care of plotting the PSD versus the appropriate frequency
    range (based on :attr:`sampling`, :attr:`NFFT` and :attr:`sides`)

    .. note:: the modification of some attributes (e.g., NFFT), makes the PSD
     obsolete. In such cases, the PSD must be re-computed before using :meth:`plot` again.

    At any time, you can get general information about the Spectrum instance::

        >>> p = Spectrum(marple_data)
        >>> print(p)
        Spectrum summary
            Data length is 64
            PSD not yet computed
            Sampling 1.0
            freq resolution 0.015625
            datatype is complex
            sides is twosided
            scal_by_freq is True

    """
    _detrend_choices = [None, 'mean']
    _sides_choices = ['onesided','twosided', 'centerdc', 'default']

    def __init__(self, data, data_y=None, sampling=1.,
                 detrend=None, scale_by_freq=True, NFFT=None):
        """**Constructor**

        .. rubric:: Attributes:

        From the input parameters, the following attributes are set:

          * :attr:`data` (updates :attr:`N`, :attr:`df`, :attr:`datatype`)
          * :attr:`data_y` used for cross PSD only (correlogram)
          * :attr:`detrend`
          * :attr:`sampling` (updates :attr:`df`)
          * :attr:`scale_by_freq`
          * :attr:`NFFT` (reset :attr:`sides`, :attr:`df`)

        The following read-only attributes are set during the initialisation:

          * :attr:`datatype`
          * :attr:`df`
          * :attr:`N`

        And finally, additional read-write attributes are available:

          * :attr:`psd`: used to store the PSD data array, which size depends
            on :attr:`sides` i.e., one-sided for real data and two-sided for
            the complex data.
          * :attr:`sides`: if set, changed the :attr:`psd`.

        """
        # user attributes. No need to bother about initialisation here
        self.__data = None
        self.__data_y = None
        self.__sampling = None
        self.__detrend = None
        self.__scale_by_freq = None
        # other attributes
        self.__sides = None
        self.__N = None
        self.__NFFT = None
        self.__df = None
        self.__datatype = None
        self.__psd = None
        self.__method = None

        # Real initialise is made here
        self.data = data
        if data_y is not None:
            self.data_y = data_y
        self.sampling = sampling
        self.sides = 'default'
        # self._range.N is the NFFT value. By default, it is the size of the data

        self._range = Range(self.__data.size, sampling) # can be private.
        self.modified = True
        self.sampling = sampling
        self.scale_by_freq = scale_by_freq

        self.NFFT = NFFT
        self.method = self.__class__  # alias to the class name

    def _getMethod(self):
        return self.__method
    def _setMethod(self, method):
        self.__method = method
    method = property(fget=_getMethod, fset=_setMethod)

    """def __call__(self, *args, **kargs):
        # To be use with care. THis function is there just to help, it
        #    does not populate the proper attribute except psd.
        if self.method is not None:
            res = self.method(self.data, *args, **kargs)
            self.psd = res[0]
        #return res
    """
    def run(self, *args, **kargs):
        # Kept for back compatibility
        # e.g. used in this external notebook:
        # http://nbviewer.jupyter.org/gist/juhasch/5182528
        self()

    def _getDetrend(self):
        return self.__detrend
    def _setDetrend(self, detrend):
        if detrend == self.__detrend:
            return
        if detrend not in self._detrend_choices:
            raise errors.SpectrumChoiceError(detrend, self._detrend_choices)
        self.__detrend = detrend
        self.modified = True
    detrend = property(fget=_getDetrend, fset=_setDetrend,
                       doc="""Getter/Setter to detrend:

                    * None: do not perform any detrend.
                    * 'mean': remove the mean value of each segment from each segment of the data.
                    * 'long-mean': remove the mean value from the data before splitting it into segments.
                    * 'linear': remove linear trend from each segment.""")

    def _get_range(self):
        return self._range
    range = property(fget=_get_range,
            doc="""Read only attribute to a :class:`Range` object.""")

    def _getScale(self):
        return self.__scale_by_freq
    def _setScale(self, scale):
        if scale == self.__scale_by_freq: 
            return
        assert scale in [True, False]
        self.__scale_by_freq = scale
        self.modified = True
    scale_by_freq = property(fget=_getScale, fset=_setScale,
                            doc="scale the PSD by :math:`2*\pi/df`")

    def _getNFFT(self):
        return self.__NFFT
    def _setNFFT(self, NFFT):#if NFFT is changed, we need to redo the padding
        if self.__NFFT == NFFT and self.__NFFT is not None:
            logging.debug('NFFT is the same, nothing to do')
            return
        new_nfft = None
        if NFFT == 'nextpow2':
            n = nextpow2(self.data.size)
            new_nfft = int(pow(2, n))
            logging.debug('NFFT is based on nextpow2: %s' % new_nfft)
        elif NFFT is None:
            logging.debug('NFFT set to data length')
            new_nfft = self.N
        elif isinstance(NFFT, int):
            logging.debug('NNFT set  manually to {}'.format(NFFT))
            assert NFFT > 0, 'NFFT must be a positive integer'
            new_nfft = NFFT
        else:
            raise ValueError("NFFT must be either None, positive integer or 'nextpow2'")

        #print new_nfft
        if self.__NFFT != new_nfft:
            self.__NFFT = new_nfft
            # Now that the NFFT has changed, we need to update the range
            self._range.N = self.__NFFT
            self.__sides = self._default_sides()
            self.modified = True
    NFFT = property(fget=_getNFFT, fset=_setNFFT, doc=
                    """Getter/Setter to the NFFT attribute.

                    :param NFFT: a user choice for setting :attr:`NFFT`.

                        * if None, the NFFT is set to :attr:`N`, the data length.
                        * if 'nextpow2', computes the next power of 2 greater
                          than or equal to the data length.
                        * if a integer is provided, it must be positive

                    If NFFT is changed, :attr:`sides` is reset and :attr:`df` as well.
                    """)

    def _default_sides(self):
        if self.datatype == 'real':
            sides = 'onesided'
        else:
            sides = 'twosided'
        return sides

    def _getSides(self):
        return self.__sides
    def _setSides(self, sides):
        # check validity of sides
        if sides not in self._sides_choices:
            raise errors.SpectrumChoiceError(sides, self._sides_choices)

        # default value
        if sides == 'default':
            sides = self._default_sides()

        # check validity of sides
        if self.datatype == 'complex':
            assert sides != ['onesided'], "complex data cannot be onesided (%s provided)" % sides

        # If sides is indeed different, update the psd
        if self.__psd is not None:
            newpsd = self.get_converted_psd(sides)
            self.__psd = newpsd
        self.__sides = sides
        logging.debug('------------> %s %s' % (self.__sides, sides))
        # we set the PSD by hand, so we can consider that PSD is up-to-date
        self.modified = False
    _doc_sides = """Getter/Setter to the :attr:`sides` attributes.

    It can be 'onesided', 'twosided', 'centerdc'. This setter changes
    :attr:`psd` to reflect the user argument choice.

    If the datatype is complex, sides cannot be one-sided.
    """
    sides = property(fget=_getSides, fset=_setSides, doc=_doc_sides)

    def _get_data_y(self):
        return self.__data_y
    def _set_data_y(self, data):
        self.__data_y = data
        self.modified = True
    data_y = property(fget=_get_data_y, fset=_set_data_y,
                      doc="""Getter/Setter to the Y-data""")

    def _getData(self):
        return self.__data
    def _setData(self, data):
        if type(data) == list:
            from numpy import array
            self.__data = array(data)
        else:
            self.__data = data.copy()
        self.__N = self.data.size # N has no setter, so we use the private version
        self.modified = True

        if numpy.isrealobj(self.__data):
            self.__datatype = 'real'
        else:
            self.__datatype = 'complex'
    data = property(fget=_getData, fset=_setData, doc="""Getter/Setter for the
    input data. If input is a list, it is cast into a numpy.array. Then, :attr:`N`,
    :attr:`df` and  :attr:`datatype` are updated.""")

    def _getPSD(self):
        if self.__psd is None or self.modified is True:
            logging.debug('Computing PSD.')
            self()
            self.modified = False
        return self.__psd
    def _setPSD(self, psd):
        # Reset the sides attribute depending on the datatype
        # if a user sets the PSD manually, there is an ambiguity because
        # input data may be odd or even length for the one-sided case
        # A twosided PSD must be even though.
        # note dec 2018: onesided can be even. e.g. if data length is N=14,
        # then psd length is 8 (N/2+1). We should introduce a more robust way
        # of setting PSD or forbid it.
        if self.datatype == 'real':
            # if psd is set we assume that the original data was even
            # so that length original data is 2 + 2*(len(psd)-1)
            #assert len(psd) % 2 == 1, 'odd data, so PSD must be one-sided'
            self.__sides = 'onesided'
            # we cannot know the NFFT used by the user when creating the 
            # PSD because depends on even or odd data. Same issue with range
            #self.__NFFT = 2 + (len(psd)-2) * 2
            self.__psd = numpy.array(psd)
            #self._range.N = self.__NFFT
        else:
            #assert len(psd) % 2 == 0, 'even  data, so PSD should be in two-sided format'
            self.__sides = 'twosided'
            self.__NFFT = len(psd)
            self.__psd = numpy.array(psd)
            self._range.N = self.__NFFT
        # we set the psd manually, so :attr:modified is reset
        self.modified = False
    psd = property(fget=_getPSD, fset=_setPSD, doc="""Getter/Setter to :attr:`psd`

    :param array psd: the array must be in agreement with the onesided/twosided
        convention: if the data in real, the psd must be onesided. If the
        data is complex, the psd must be twosided.

    When you set this attribute, several attributes are set:

        * :attr:`sides` is set to onesided if datatype is real and twosided
          if datatype is complex.
        * :attr:`NFFT` is set to len(psd)*2 if sides=onesided and (len(psd)-1)*2
          if sides=twosided.
        * :attr:`range`.N is set to NFFT, which update :attr:`df`.

    """)

    # simple getter/setter for the N attribute. Updates df
    def _getN(self):
        return self.__N
    N = property(fget=_getN,
                 doc="""Getter to the original data size. :attr:`N` is automatically
                 updated when changing the data only.""")

    #simple getter/setter for the sampling attribute. Updates sampling and df
    def _getSampling(self):
        return self.__sampling
    def _setSampling(self, sampling):
        if sampling == self.__sampling: return
        self.__sampling = sampling
        self.__df = self.__sampling / float(self.__N)
        self.modified = True
    sampling = property(fget=_getSampling, fset=_setSampling,
        doc="""Getter/Setter to sampling frequency. Updates the :attr:`df` automatically.""")

    #the frequency resolution. Read only.
    def _getdf(self):
        return self._range.df
    df = property(fget=_getdf, doc="""Getter to step frequency. This attribute
 is updated as soon as :attr:`data` or :attr:`sampling` is changed""")

    def _getdatatype(self):
        return self.__datatype
    datatype = property(fget=_getdatatype,
                    doc="""Getter to the datatype ('real' or 'complex').
                    :attr:`datatype` is automatically updated when changing
                    the data.""")

    def scale(self):
        if self.scale_by_freq is True:
            self.psd *= 2*numpy.pi/self.df

    def frequencies(self, sides=None):

        """Return the frequency vector according to :attr:`sides`"""
        # use the attribute sides except if a valid sides argument is provided
        if sides is None:
            sides = self.sides
        if sides not in self._sides_choices:
            raise errors.SpectrumChoiceError(sides, self._sides_choices)

        if sides == 'onesided':
            return self._range.onesided()
        if sides == 'twosided':
            return self._range.twosided()
        if sides == 'centerdc':
            return self._range.centerdc()

    def get_converted_psd(self, sides):
        """This function returns the PSD in the **sides** format

        :param str sides: the PSD format in ['onesided', 'twosided', 'centerdc']
        :return: the expected PSD.

        .. doctest::

            from spectrum import *
            p = pcovar(marple_data, 15)
            centerdc_psd = p.get_converted_psd('centerdc')

        .. note:: this function does not change the object, in particular, it
            does not change the :attr:`psd` attribute. If you want to change
            the psd on the fly, change the attribute :attr:`sides`.

        """
        if sides == self.sides:
            #nothing to be done is sides = :attr:`sides
            return self.__psd

        if self.datatype == 'complex':
            assert sides != 'onesided', \
                "complex datatype so sides cannot be onesided."

        if self.sides == 'onesided':
            logging.debug('Current sides is onesided')
            if sides == 'twosided':
                logging.debug('--->Converting to twosided')
                # here we divide everything by 2 to get the twosided version
                #N = self.NFFT
                newpsd = numpy.concatenate((self.psd[0:-1]/2., list(reversed(self.psd[0:-1]/2.))))
                # so we need to multiply by 2 the 0 and FS/2 frequencies
                newpsd[-1] = self.psd[-1]
                newpsd[0] *= 2.
            elif sides == 'centerdc':
                # FIXME. this assumes data is even so PSD is stored as
                # P0 X1 X2 X3 P1
                logging.debug('--->Converting to centerdc')
                P0 = self.psd[0]
                P1 = self.psd[-1]
                newpsd = numpy.concatenate((self.psd[-1:0:-1]/2., self.psd[0:-1]/2.))
                # so we need to multiply by 2 the 0 and F2/2 frequencies
                #newpsd[-1] = P0 / 2
                newpsd[0] = P1
        elif self.sides == 'twosided':
            logging.debug('Current sides is twosided')
            if sides == 'onesided':
                # we assume that data is stored as X0,X1,X2,X3,XN
                # that is original data is even.
                logging.debug('Converting to onesided assuming ori data is even')
                midN = (len(self.psd)-2) / 2
                newpsd = numpy.array(self.psd[0:int(midN)+2]*2)
                newpsd[0] /= 2
                newpsd[-1] = self.psd[-1]
            elif sides == 'centerdc':
                newpsd = stools.twosided_2_centerdc(self.psd)
        elif self.sides == 'centerdc': # same as twosided to onesided
            logging.debug('Current sides is centerdc')
            if sides == 'onesided':
                logging.debug('--->Converting to onesided')
                midN = int(len(self.psd) / 2)
                P1 = self.psd[0]
                newpsd = numpy.append(self.psd[midN:]*2, P1)
            elif sides == 'twosided':
                newpsd = stools.centerdc_2_twosided(self.psd)
        else:
            raise ValueError("sides must be set to 'onesided', 'twosided' or 'centerdc'")

        return newpsd

    def plot(self, filename=None, norm=False, ylim=None,
              sides=None,  **kargs):
        """a simple plotting routine to plot the PSD versus frequency.

        :param str filename: save the figure into a file
        :param norm: False by default. If True, the PSD is normalised.
        :param ylim: readjust the y range .
        :param sides: if not provided, :attr:`sides` is used. See :attr:`sides`
            for details.
        :param kargs: any optional argument accepted by :func:`pylab.plot`.

        .. plot::
            :width: 80%
            :include-source:

            from spectrum import *
            p = Periodogram(marple_data)
            p.plot(norm=True, marker='o')

        """
        import pylab
        from pylab import ylim as plt_ylim
        #First, check that psd attribute is up-to-date
        # just to get the PSD to be recomputed if needed
        _ = self.psd


        # check that the input sides parameter is correct if provided
        if sides is not None:
            if sides not in self._sides_choices:
                raise errors.SpectrumChoiceError(sides, self._sides_choices)

        # if sides is provided but identical to the current psd, nothing to do.
        # if sides not provided, let us use self.sides
        if sides is None or sides == self.sides:
            frequencies = self.frequencies()
            psd = self.psd
            sides = self.sides
        elif sides is not None:
            # if sides argument is different from the attribute, we need to
            # create a new PSD/Freq ; indeed we do not want to change the
            # attribute itself

            # if data is complex, one-sided is wrong in any case.
            if self.datatype == 'complex':
                if sides == 'onesided':
                    raise ValueError("sides cannot be one-sided with complex data")

            logging.debug("sides is different from the one provided. Converting PSD")
            frequencies = self.frequencies(sides=sides)
            psd = self.get_converted_psd(sides)

        if len(psd) != len(frequencies):
            raise ValueError("PSD length is %s and freq length is %s" % (len(psd), len(frequencies)))

        if 'ax' in list(kargs.keys()):
            save_ax = pylab.gca()
            pylab.sca(kargs['ax'])
            rollback = True
            del kargs['ax']
        else:
            rollback = False

        if norm:
            pylab.plot(frequencies, 10 * stools.log10(psd/max(psd)),  **kargs)
        else:
            pylab.plot(frequencies, 10 * stools.log10(psd),**kargs)

        pylab.xlabel('Frequency')
        pylab.ylabel('Power (dB)')
        pylab.grid(True)

        if ylim:
            plt_ylim(ylim)

        if sides == 'onesided':
            pylab.xlim(0, self.sampling/2.)
        elif sides == 'twosided':
            pylab.xlim(0, self.sampling)
        elif sides == 'centerdc':
            pylab.xlim(-self.sampling/2., self.sampling/2.)

        if filename:
            pylab.savefig(filename)

        if rollback:
            pylab.sca(save_ax)

        del psd, frequencies #is it needed?

    def power(self):
        r"""Return the power contained in the PSD

        if scale_by_freq is False, the power is:

        .. math:: P = N \sum_{k=1}^{N} P_{xx}(k)

        else, it is

        .. math:: P =  \sum_{k=1}^{N} P_{xx}(k) \frac{df}{2\pi}

        .. todo:: check these equations


        """
        if self.scale_by_freq == False:
            return sum(self.psd) * len(self.psd)
        else:
            return sum(self.psd) * self.df/(2.*numpy.pi)

    def _str_title(self):
        return "Spectrum summary\n"

    def __str__(self):
        msg = self._str_title()
        msg += "    Data length is %s\n" % self.data.size
        try:
            msg += "    PSD length is %s\n" % self.psd.size
        except:
            msg += "    PSD not yet computed\n"
        msg += "    Sampling %s\n" % self.sampling
        msg += "    freq resolution %s\n" % self.df
        msg += "    datatype is %s\n" % self.datatype
        msg += "    sides is %s\n" % self.sides
        msg += "    scal_by_freq is %s\n" % self.scale_by_freq
        return msg


class ParametricSpectrum(Spectrum):
    """Spectrum based on Fourier transform.

    This class inherits attributes and methods from
    :class:`Spectrum`. It is used by children class :class:`~spectrum.periodogram.Periodogram`,
    :class:`~spectrum.correlog.pcorrelogram` and :class:`Welch` PSD estimates.
    The parameters are those used by :class:`Spectrum`.

    :param array data:     Input data (list or numpy.array)
    :param float sampling: sampling frequency of the input :attr:`data`
    :param str detrend:    detrend method ([None,'mean']) to apply on the input
        data before computing the PSD. See :attr:`detrend`.
    :param bool scale_by_freq: Divide the final PSD by :math:`2*\pi/df`

    In addition you need specific parameters such as:

    :param str window:  a tapering window. See :class:`~spectrum.window.Window`.
    :param int lag:     to be used by the :class:`~spectrum.correlog.pcorrelogram` methods only.
    :param int NFFT:    Total length of the data given to the FFT

    This class has dedicated PSDs methods such as :meth:`periodogram`, which
    are equivalent to children class such as :class:`~spectrum.periodogram.Periodogram`.

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import datasets
        from spectrum import ParametricSpectrum
        data = datasets.data_cosine(N=1024)
        s = ParametricSpectrum(data, ar_order=4, ma_order=4, sampling=1024, NFFT=512, lag=10)
        #s.parma()
        #s.plot(sides='onesided')
        #s.plot(sides='twosided')

    """
    def __init__(self, data, sampling=1., ar_order=None, ma_order=None,
            lag=-1, NFFT=None, detrend=None, scale_by_freq=True):
        """**Constructor**

        See the class documentation for the parameters.

        .. rubric:: Additional attributes to those inherited
            from :class:`Spectrum`:

        * :attr:`ar_order`, the ar order of the PSD estimates
        * :attr:`ma_order`, the ar order of the PSD estimates

        """
        super(ParametricSpectrum, self).__init__(data, sampling=sampling,
                                                 NFFT=NFFT,
                                                 scale_by_freq=scale_by_freq,
                                                 detrend=detrend)
        if ar_order is None and ma_order is None:
            raise errors.SpectrumARMAError
        #new user attributes
        self.__ar_order = ar_order
        self.__ma_order = ma_order
        self.ar_order = ar_order
        self.ma_order = ma_order
        self.lag = lag
        # will be populated when running an ARMA PSD estimate
        self.__ar = None
        self.__ma = None
        self.__reflection = None
        self.__rho = None

    def _set_ar_order(self, ar):
        if ar is not None:
            if ar < 0:
                raise errors.SpectrumARError
            self.__ar_order = ar
    def _get_ar_order(self):
        return self.__ar_order
    ar_order = property(fget=_get_ar_order, fset=_set_ar_order, doc="")

    def _set_ma_order(self, ma):
        if ma is not None:
            if ma < 0:
                raise errors.SpectrumMAError
            self.__ma_order = ma
        else:
            self.__ma_order = None
    def _get_ma_order(self):
        return self.__ma_order
    ma_order = property(fget=_get_ma_order, fset=_set_ma_order, doc="")

    def _set_ma(self, ma):
        self.__ma = ma
    def _get_ma(self):
        return self.__ma
    ma = property(fget=_get_ma, fset=_set_ma, doc="")

    def _set_ar(self, ar):
        self.__ar = ar
    def _get_ar(self):
        return self.__ar
    ar = property(fget=_get_ar, fset=_set_ar, doc="")

    def _set_rho(self, rho):
        self.__rho = rho
    def _get_rho(self):
        return self.__rho
    rho = property(fget=_get_rho, fset=_set_rho)

    def _set_ref(self, ref):
        self.__reflection = ref
    def _get_ref(self):
        return self.__reflection
    reflection = property(fget=_get_ref, fset=_set_ref)

    """ self.reflection = None
        elif method == 'aryule':
            from spectrum import aryule
            ar, v, coeff = aryule(self.data, self.ar_order)
            self.ar = ar
            self.rho = v
            self.reflection = coeff
        """

    def plot_reflection(self):
        from pylab import stem, title, xlabel, ylabel
        if self.reflection is not None:
            stem(list(range(0, len(self.reflection))), abs(self.reflection))
            title('Reflection coefficient evolution')
            xlabel('Order')
            ylabel('Reflection Coefficient absolute values')
        else:
            logging.warning("Reflection coefficients not available or not yet computed.")

    def _str_title(self):
        return "ParametricSpectrum summary\n"

    def __str__(self):
        return super(ParametricSpectrum, self).__str__()



class FourierSpectrum(Spectrum):
    """Spectrum based on Fourier transform.

    This class inherits attributes and methods from  :class:`Spectrum`. It is
    used by children class :class:`~spectrum.periodogram.Periodogram`,
    :class:`~spectrum.correlog.pcorrelogram` and :class:`Welch` PSD estimates.

    The parameters are those used by :class:`Spectrum`

    :param array data:     Input data (list or numpy.array)
    :param data_y: input data required to perform cross-PSD only.
    :param float sampling: sampling frequency of the input :attr:`data`
    :param str detrend:    detrend method ([None,'mean']) to apply on the input
        data before computing the PSD. See :attr:`detrend`.
    :param bool scale_by_freq: Divide the final PSD by :math:`2*\pi/df`
    :param int NFFT: total length of the data given to the FFT

    In addition you need specific parameters such as:

    :param str window:  a tapering window. See :class:`~spectrum.window.Window`.
    :param int lag:     to be used by the :class:`~spectrum.correlog.pcorrelogram` methods only.


    This class has dedicated PSDs methods such as :meth:`speriodogram`, which
    are equivalent to children class such as :class:`~spectrum.periodogram.Periodogram`.

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import datasets
        from spectrum import FourierSpectrum
        s = FourierSpectrum(datasets.data_cosine(), lag=32, sampling=1024, NFFT=512)
        s.periodogram()
        s.plot(label='periodogram')
        #s.correlogram()
        s.plot(label='correlogram')
        from pylab import legend, xlim
        legend()
        xlim(0,512)

    """
    _window = window_names

    def __init__(self, data, sampling=1.,
                 window='hanning', NFFT=None, detrend=None,
                 scale_by_freq=True, lag=-1):
        """**Constructor**

        See the class documentation for the parameters.

        .. rubric:: Additional attributes to those inherited from

        :class:`Spectrum` are:

        * :attr:`lag`, a lag used to compute the autocorrelation
        * :attr:`window`, the tapering window to be used

        """
        super(FourierSpectrum, self).__init__(data,
                sampling=sampling, detrend=detrend,
                scale_by_freq=scale_by_freq, NFFT=NFFT)
        self.__window = None
        self.__lag = None
        #self.P0 = None
        self.lag = lag
        self.window = window

    def _set_window(self, window):
        if window == self.__window:
            return
        if window not in self._window:
            raise errors.SpectrumChoiceError(window, self._window)
        self.__window = window
        self.modified = True
    def _get_window(self):
        return self.__window
    window = property(fget=_get_window, fset=_set_window,
                      doc="""Tapering window to be applied""")

    def _set_lag(self, lag):
        if lag == self.__lag:
            return
        self.__lag = lag
        self.modified = True
    def _get_lag(self):
        return self.__lag
    lag = property(fget=_get_lag, fset=_set_lag, doc="""Getter/Setter used by the correlogram when
        autocorrelation estimates are required.""")

    def _str_title(self):
        return "FourierSpectrum summary\n"

    def periodogram(self):
        """An alias to :class:`~spectrum.periodogram.Periodogram`

        The parameters are extracted from the attributes. Relevant attributes
        ares :attr:`window`, attr:`sampling`, attr:`NFFT`, attr:`scale_by_freq`,
        :attr:`detrend`.

        .. plot::
            :width: 80%
            :include-source:

            from spectrum import datasets
            from spectrum import FourierSpectrum
            s = FourierSpectrum(datasets.data_cosine(), sampling=1024, NFFT=512)
            s.periodogram()
            s.plot()
        """
        from .periodogram import speriodogram
        psd = speriodogram(self.data, window=self.window, sampling=self.sampling,
                             NFFT=self.NFFT, scale_by_freq=self.scale_by_freq,
                             detrend=self.detrend)
        self.psd = psd





import logging

import numpy as np
from .psd import ParametricSpectrum


__all__ = ["arcovar", "arcovar_marple", "pcovar", 'arcovar_marple']


def arcovar_marple(x, order):
    r"""Estimate AR model parameters using covariance method

    This implementation is based on [Marple]_. This code is far more
    complicated and slower than :func:`arcovar` function, which is now the official version.
    See :func:`arcovar` for a detailed description of Covariance method.

    This function should be used in place of arcovar only if order<=4, for
    which :func:`arcovar` does not work.

    Fast algorithm for the solution of the covariance least squares normal
    equations from Marple.

    :param array X:  Array of complex data samples
    :param int oder: Order of linear prediction model

    :return:
        * AF   - Array of complex forward linear prediction coefficients
        * PF   - Real forward linear prediction variance at order IP
        * AB   - Array of complex backward linear prediction coefficients
        * PB   - Real backward linear prediction variance at order IP
        * PV   - store linear prediction coefficients

    .. note:: this code and the original code in Marple diverge for ip>10.
        it seems that this is related to single precision used with
        complex type in fortran whereas numpy uses double precision for
        complex type.

    :validation: the AR parameters are the same as those returned by
        a completely different function :func:`arcovar`.

    :References: [Marple]_
    """
    assert len(x) >= order, "X must be dimensioned >=N"

    #   ----------------------------------------------------- Initialization
    x = np.array(x)
    N = len(x)


    # Equations 8.C.42
    r0 = sum(abs(x)**2.)
    r1 = abs(x[0])**2
    rN = abs(x[N-1])**2

    pf = r0 - r1
    pb = r0 - rN

    delta = 1. - r1 / r0
    gamma = 1. - rN / r0

    c = np.zeros(N, dtype=complex)
    d = np.zeros(N, dtype=complex)
    r = np.zeros(N, dtype=complex)
    af = np.zeros(N, dtype=complex)
    ab = np.zeros(N, dtype=complex)

    c[0] = x[N-1].conjugate() / r0
    d[0] = x[0].conjugate() / r0

    # special case
    if order == 0:
        pf = r0 / float(N)
        pb = pf
        return af, pf, ab, pb, 0

    # ----------------------------------------------------------  MAIN LOOP

    #ip +1 because we want to enter in the loop to run the first part of the code.
    pbv = []
    for m in range(0, order+1):
        logging.debug('----------------------------m={}'.format(m))
        logging.debug(c[0:2])
        logging.debug(d[0:2])
        r1 = 1./pf
        r2 = 1./pb
        r3 = 1./delta
        r4 = 1./gamma

        #logging.debug('starting r1r2r3r4=', r1, r2, r3, r4, pf, pb, delta, gamma)

        #Order update: AF and AB vectors ; time update: C and D vectors
        temp = 0.+0.j
        for k in range(m+1, N):
            temp = temp + x[k]*x[k-m-1].conjugate()

        r[m] =  temp.conjugate()
        theta = x[0] * c[m]
        #print(('theta', theta))
        #    print(('c=', c[0:2]))
        #    print(('d=', d[0:2]))
        if m == 0:
            pass
        else:
            for k in range(0, m):
                theta = theta + x[m-k] * c[k]                   # Eq. (8.C.39)
                r[k] = r[k] - x[N-m-1] * x[N-m+k].conjugate() # Eq. (8.C.32)
                temp = temp + af[m-k-1] * r[k].conjugate()
                #print 'loop1 k=', k
                #print '         theta=',theta, 'r[k]=',r[k], 'temp=', temp
                #print '         c=',c[k], 'af=',af[m-k-1]
        """if m > 0:
            if debug:
                print((m, N-m))
                print(('Xk=0',x[m-0],x[N-m-1], x[N-m+0]))
            if m > 1:
                if debug:
                    print('Xk=1',x[m-1],x[N-m-1], x[N-m+1])
        """
        c1 = -temp * r2
        c2 = -r1 * temp.conjugate()
        c3 = theta * r3
        c4 = r4 *theta.conjugate()
        #if debug:
        #    print('c1 c2 c3 c4 before af=',c1 ,c2 ,c3 ,c4)
        af[m] = c1                    #             ! Eq. (8.C.19)
        ab[m] = c2                    #             ! Eq. (8.C.22)
        save = c[m]
        c[m] = save + c3*d[m]
        d[m] = d[m] + c4*save
        #if debug:
        #    print('res',m,'af[m]=',af[m], ab[m], save, 'temp=',temp)

        if m == 0:
            pass
        else:
            #if debug:print('af before', af[0:2])
            for k in range(0, m):
                save = af[k]
                af[k] = save + c1 * ab[m-k-1] # Eq. (8.C.18)
                ab[m-k-1] = ab[m-k-1] + c2 * save   # Eq. (8.C.21)

                save = c[k]
                c[k] = save + c3*d[k]       # Eq. (8.C.37)
                d[k] = d[k] + c4*save       # Eq. (8.C.38)
                #if debug:
                #    print('loop2 k=', k)
                #    print('      af[k]=', af[k])
                #    print('      ab[m-k-1]=', ab[m-k-1])
                #    print('      c[k]=', c[k])
                #    print('      d[k]=', d[k])
            #if debug:
            #    print('af after=', af[0:2])
            #    print('ab=', ab[0:2])

        r5 = temp.real**2 + temp.imag**2
        pf = pf - r5*r2         # Eq. (8.C.20)
        pb = pb - r5*r1         # Eq. (8.C.23)
        r5 = theta.real**2 + theta.imag**2
        delta = delta - r5*r4               # Eq. (8.C.39)
        gamma = gamma - r5*r3               # Eq. (8.C.40)
        #if debug:
        #    print('r5r2r1deltagamma', r5, r2, r1 , delta, gamma)
        #    print('pf before norm', pf, pb, N-m-1)
        if m != order-1:
            pass
        else:
            pf = pf / float(N-m-1)
            pb = pb / float(N-m-1)
            #if debug:
            #    print('ENDING', N-m-1)
            break

        #if debug:
        #    print('pf and pb', pf, pb)
        if pf > 0 and pb > 0:
            pass
        else:
            ValueError("Negative PF or PB value")
        if (delta > 0. and delta <=1 and gamma > 0. and gamma <=1):
            pass
        else:
            ValueError("Invalid delta or gamma value")

        #C   Time update:  AF and AB vectors; order update:  C and D vectors

        r1 = 1./pf
        r2 = 1./pb
        r3 = 1./delta
        r4 = 1./gamma
        #if debug:
        #    print('--------time update', r1, r2, r3, r4, m+1, N-m-1, x[m+1], x[N-m-2])
        ef = x[m+1]
        eb = x[(N-1)-m-1]


        for k in range(0,m+1):
            #print 'k=', k, 'ef=', ef, ' eb=',eb,' af=',af[k], ab[k]
            #print x[m-k],x[N-m+k-1]
            ef = ef + af[k] * x[m-k]             # Eq. (8.C.1)
            eb = eb + ab[k] * x[N-m+k-1]                   # Eq. (8.C.2)

        #ef = sum(af)

        #if debug:
        #    print('efweb', ef , eb)
        c1 = ef*r3
        c2 = eb*r4
        c3 = eb.conjugate() * r2
        c4 = ef.conjugate() * r1
        #if debug:
        #    print('c1c2c3c4', c1, c2, c3, c4)
        #    print('af before', af[0:2])
        for k in range(m, -1, -1):
            save = af[k]
            af[k] = save + c1 * d[k]                    #  Eq. (8.C.33)
            d[k+1] = d[k] + c4 * save                    # Eq. (8.C.25)
            save = ab[k]
            ab[k] = save + c2 * c[m-k]                 # Eq. (8.C.35)
            c[m-k] = c[m-k] + c3 * save              # Eq. (8.C.24)
        #if debug:
        #    print('af after', af[0:2])
        #    print('d', d[0:2])
        #    print('ab', ab[0:2])
        #    print('c', c[0:2])
        #if debug:print('Pb before', pf, pb)
        c[m+1] = c3
        d[0] = c4
        #r5 = abs(ef)**2
        r5 = ef.real**2 + ef.imag**2
        pf = pf - r5 * r3                              # Eq. (8.C.34)

        delta = delta-r5 * r1                        # Eq. (8.C.30)
        #r5 = abs(eb)**2
        r5 = eb.real**2 + eb.imag**2
        pb = pb - r5 * r4                              # Eq. (8.C.36)
        #if debug:
        #    print('Pb---------------------', m, pb, r5, r4)
        gamma = gamma-r5*r2                        # Eq. (8.C.31)
        pbv.append(pb)

        if (pf > 0. and pb > 0.):
            pass
        else:
            ValueError("Negative PF or PB value")
        #if debug:
        #    print(delta, gamma)
        if (delta > 0. and delta <= 1.) and (gamma > 0. and gamma <= 1.):
            pass
        else:
            ValueError("Invalid delta or gamma value")


    #af=array of forward coeff
    #ab=array of barward coeff
    #pb=backward variance
    #pf=forward variance
    return af, pf, ab, pb, pbv


def arcovar(x, order):
    r"""Simple and fast implementation of the covariance AR estimate

    This code is 10 times faster than :func:`arcovar_marple` and more importantly
    only 10 lines of code, compared to a 200 loc for :func:`arcovar_marple`


    :param array X:  Array of complex data samples
    :param int oder: Order of linear prediction model

    :return:
        * a - Array of complex forward linear prediction coefficients
        * e - error

    The covariance method fits a Pth order autoregressive (AR) model to the
    input signal, which is assumed to be the output of
    an AR system driven by white noise. This method minimizes the forward
    prediction error in the least-squares sense. The output vector
    contains the normalized estimate of the AR system parameters

    The white noise input variance estimate is also returned.

    If is the power spectral density of y(n), then:

    .. math:: \frac{e}{\left| A(e^{jw}) \right|^2} = \frac{e}{\left| 1+\sum_{k-1}^P a(k)e^{-jwk}\right|^2}

    Because the method characterizes the input data using an all-pole model,
    the correct choice of the model order p is important.

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import arcovar, marple_data, arma2psd
        from pylab import plot, log10, linspace, axis

        ar_values, error = arcovar(marple_data, 15)
        psd = arma2psd(ar_values, sides='centerdc')
        plot(linspace(-0.5, 0.5, len(psd)), 10*log10(psd/max(psd)))
        axis([-0.5, 0.5, -60, 0])

    .. seealso:: :class:`pcovar`

    :validation: the AR parameters are the same as those returned by
        a completely different function :func:`arcovar_marple`.

    :References: [Mathworks]_
    """

    from spectrum import corrmtx
    import scipy.linalg

    X = corrmtx(x, order, 'covariance')
    Xc = np.array(X[:, 1:])
    X1 = np.array(X[:, 0])

    # Coefficients estimated via the covariance method
    # Here we use lstsq rathre than solve function because Xc is not square
    # matrix

    a, _residues, _rank, _singular_values = scipy.linalg.lstsq(-Xc, X1)

    # Estimate the input white noise variance
    Cz = np.dot(X1.conj().transpose(), Xc)
    e = np.dot(X1.conj().transpose(), X1) + np.dot(Cz, a)
    assert e.imag < 1e-4, 'wierd behaviour'
    e = float(e.real) # ignore imag part that should be small

    return a, e


class pcovar(ParametricSpectrum):
    """Class to create PSD based on covariance algorithm

    See :func:`arcovar` for description.

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import pcovar, marple_data
        p = pcovar(marple_data, 15, NFFT=4096)
        p.plot(sides='centerdc')

    .. seealso:: :class:`arcovar`
    """
    def __init__(self, data, order, NFFT=None, sampling=1.,
                 scale_by_freq=False):
        """**Constructor**

        For a detailled description of the parameters, see :func:`arcovar`.

        :param array data: input data (list or numpy.array)
        :param int order:
        :param int NFFT: total length of the final data sets (padded
            with zero if needed; default is 4096)
        :param float sampling: sampling frequency of the input :attr:`data`.

        """
        super(pcovar, self).__init__(data, ar_order=order,
                                            NFFT=NFFT, sampling=sampling,
                                            scale_by_freq=scale_by_freq)

    def __call__(self):
        from spectrum import arma2psd
        ar, _e = arcovar(self.data, self.ar_order)
        self.ar = ar
        psd = arma2psd(A=ar, T=self.sampling, NFFT=self.NFFT)

        if self.datatype == 'real':
            if self.NFFT % 2 == 0:
                newpsd = psd[0:int(self.NFFT/2+1)] * 2
            else:
                newpsd = psd[0:int((self.NFFT+1)/2)] * 2
            self.psd = newpsd
        else:
            self.psd = psd
        if self.scale_by_freq is True:
            self.scale()
        return self

    def _str_title(self):
        return "Covariance PSD estimate\n"

    def __str__(self):
        return super(pcovar, self).__str__()





import numpy
from numpy import pi

__all__ = ['morlet', 'chirp', 'mexican', 'meyeraux']


def morlet(lb, ub, n):
    r"""Generate the Morlet waveform


    The Morlet waveform is defined as follows:

    .. math:: w[x] = \cos{5x}  \exp^{-x^2/2}

    :param lb: lower bound
    :param ub: upper bound
    :param int n: waveform data samples


    .. plot::
        :include-source:
        :width: 80%

        from spectrum import morlet
        from pylab import plot
        plot(morlet(0,10,100))

    """
    if n <= 0:
        raise ValueError("n must be strictly positive")

    x = numpy.linspace(lb, ub, n)
    psi = numpy.cos(5*x) * numpy.exp(-x**2/2.)
    return psi


def chirp(t, f0=0., t1=1., f1=100., form='linear', phase=0):
    r"""Evaluate a chirp signal at time t.

    A chirp signal is a frequency swept cosine wave.

    .. math:: a = \pi  (f_1 - f_0) / t_1
    .. math:: b = 2  \pi  f_0
    .. math:: y = \cos\left( \pi\frac{f_1-f_0}{t_1}  t^2 + 2\pi f_0 t + \rm{phase} \right)

    :param array t: times at which to evaluate the chirp signal
    :param float f0: frequency at time t=0 (Hz)
    :param float t1: time t1
    :param float f1: frequency at time t=t1 (Hz)
    :param str form: shape of frequency sweep in ['linear', 'quadratic', 'logarithmic']
    :param float phase: phase shift at t=0

    The parameter **form** can be:

        * 'linear'      :math:`f(t) = (f_1-f_0)(t/t_1) + f_0`
        * 'quadratic'   :math:`f(t) = (f_1-f_0)(t/t_1)^2 + f_0`
        * 'logarithmic' :math:`f(t) = (f_1-f_0)^{(t/t_1)} + f_0`

    Example:

    .. plot::
        :include-source:
        :width: 80%

        from spectrum import chirp
        from pylab import linspace, plot
        t = linspace(0, 1, 1000)
        y = chirp(t, form='linear')
        plot(y)
        y = chirp(t, form='quadratic')
        plot(y, 'r')

    """
    valid_forms = ['linear', 'quadratic', 'logarithmic']
    if form not in valid_forms:
        raise ValueError("Invalid form. Valid form are %s"
            % valid_forms)
    t = numpy.array(t)
    phase = 2. * pi * phase / 360.
    if form == "linear":
        a = pi * (f1 - f0)/t1
        b = 2. * pi * f0
        y = numpy.cos(a * t**2 + b*t + phase)
    elif form == "quadratic":
        a = (2/3. * pi * (f1-f0)/t1/t1)
        b = 2. * pi * f0
        y = numpy.cos(a*t**3 + b * t + phase)
    elif form == "logarithmic":
        a = 2. * pi * t1/numpy.log(f1-f0)
        b = 2. * pi * f0
        x = (f1-f0)**(1./t1)
        y = numpy.cos(a * x**t + b * t + phase)

    return y


def mexican(lb, ub, n):
    r"""Generate the mexican hat wavelet

    The Mexican wavelet is:

    .. math:: w[x] = \cos{5x}  \exp^{-x^2/2}

    :param lb: lower bound
    :param ub: upper bound
    :param int n: waveform data samples
    :return: the waveform

    .. plot::
        :include-source:
        :width: 80%

        from spectrum import mexican
        from pylab import plot
        plot(mexican(0, 10, 100))

    """
    if n <= 0:
        raise ValueError("n must be strictly positive")

    x = numpy.linspace(lb, ub, n)
    psi = (1.-x**2.) * (2./(numpy.sqrt(3.)*pi**0.25)) * numpy.exp(-x**2/2.)
    return psi


def meyeraux(x):
    r"""Compute the Meyer auxiliary function

    The Meyer function is

    .. math:: y = 35 x^4-84 x^5+70 x^6-20 x^7

    :param array x:
    :return: the waveform

    .. plot::
        :include-source:
        :width: 80%

        from spectrum import meyeraux
        from pylab import linspace, plot
        t = linspace(0, 1, 1000)
        plot(t, meyeraux(t))

    """

    return  35*x**4-84.*x**5+70.*x**6-20.*x**7




"""ARMA and MA estimates, ARMA and MA PSD estimates.

.. topic:: ARMA model and Power Spectral Densities.

    .. autosummary::
       :nosignatures:

       arma_estimate
       ma
       arma2psd
       parma
       pma

    .. codeauthor:: Thomas Cokelaer 2011

    :References: See [Marple]_
"""
import numpy as np
from numpy.fft import fft

from spectrum.correlation import CORRELATION
from spectrum.covar import arcovar, arcovar_marple
import spectrum.yulewalker as yulewalker
from spectrum.psd import ParametricSpectrum


__all__ = ["arma2psd", "arma_estimate", "ma", "pma", "parma"]


def arma2psd(A=None, B=None, rho=1., T=1., NFFT=4096, sides='default',
        norm=False):
    r"""Computes power spectral density given ARMA values.

    This function computes the power spectral density values
    given the ARMA parameters of an ARMA model. It assumes that
    the driving sequence is a white noise process of zero mean and
    variance :math:`\rho_w`. The sampling frequency and noise variance are
    used to scale the PSD output, which length is set by the user with the
    `NFFT` parameter.

    :param array A:   Array of AR parameters (complex or real)
    :param array B:   Array of MA parameters (complex or real)
    :param float rho: White noise variance to scale the returned PSD
    :param float T:   Sampling frequency in Hertz to scale the PSD.
    :param int NFFT:  Final size of the PSD
    :param str sides: Default PSD is two-sided, but sides can be set to centerdc.

    .. warning:: By convention, the AR or MA arrays does not contain the
        A0=1 value.

    If :attr:`B` is None, the model is a pure AR model. If :attr:`A` is None,
    the model is a pure MA model.

    :return: two-sided PSD

    .. rubric:: Details:

    AR case: the power spectral density is:

    .. math:: P_{ARMA}(f) = T \rho_w \left|\frac{B(f)}{A(f)}\right|^2

    where:

    .. math:: A(f) = 1 + \sum_{k=1}^q b(k) e^{-j2\pi fkT}
    .. math:: B(f) = 1 + \sum_{k=1}^p a(k) e^{-j2\pi fkT}

    .. rubric:: **Example:**

    .. plot::
        :width: 80%
        :include-source:

        import spectrum.arma
        from pylab import plot, log10, legend
        plot(10*log10(spectrum.arma.arma2psd([1,0.5],[0.5,0.5])), label='ARMA(2,2)')
        plot(10*log10(spectrum.arma.arma2psd([1,0.5],None)), label='AR(2)')
        plot(10*log10(spectrum.arma.arma2psd(None,[0.5,0.5])), label='MA(2)')
        legend()

    :References: [Marple]_
    """
    if NFFT is None:
        NFFT = 4096

    if A is None and B is None:
        raise ValueError("Either AR or MA model must be provided")

    psd = np.zeros(NFFT, dtype=complex)

    if A is not None:
        ip = len(A)
        den = np.zeros(NFFT, dtype=complex)
        den[0] = 1.+0j
        for k in range(0, ip):
            den[k+1] = A[k]
        denf = fft(den, NFFT)

    if B is not None:
        iq = len(B)
        num = np.zeros(NFFT, dtype=complex)
        num[0] = 1.+0j
        for k in range(0, iq):
            num[k+1] = B[k]
        numf = fft(num, NFFT)

    # Changed in version 0.6.9 (divided by T instead of multiply)
    if A is not None and B is not None:
        psd = rho / T * abs(numf)**2. / abs(denf)**2.
    elif A is not None:
        psd = rho / T / abs(denf)**2.
    elif B is not None:
        psd = rho / T * abs(numf)**2.


    psd = np.real(psd)
    # The PSD is a twosided PSD.
    # to obtain the centerdc
    if sides != 'default':
        from . import tools
        assert sides in ['centerdc']
        if sides == 'centerdc':
            psd = tools.twosided_2_centerdc(psd)

    if norm == True:
        psd /= max(psd)

    return psd


def arma_estimate(X, P, Q, lag):
    """Autoregressive and moving average estimators.

    This function provides an estimate of the autoregressive
    parameters, the moving average parameters, and the driving
    white noise variance of  an ARMA(P,Q) for a complex or real data sequence.

    The parameters are estimated using three steps:

        * Estimate the AR parameters from the original data based on a least
          squares modified Yule-Walker technique,
        * Produce a residual time sequence by filtering the original data
          with a filter based on the AR parameters,
        * Estimate the MA parameters from the residual time sequence.

    :param array X: Array of data samples (length N)
    :param int P: Desired number of AR parameters
    :param int Q: Desired number of MA parameters
    :param int lag: Maximum lag to use for autocorrelation estimates

    :return:
        * A     - Array of complex P AR parameter estimates
        * B     - Array of complex Q MA parameter estimates
        * RHO   - White noise variance estimate

    .. note::
      *  lag must be >= Q (MA order)

    **dependencies**:
        * :meth:`spectrum.correlation.CORRELATION`
        * :meth:`spectrum.covar.arcovar`
        * :meth:`spectrum.arma.ma`

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import arma_estimate, arma2psd, marple_data
        import pylab

        a,b, rho = arma_estimate(marple_data, 15, 15, 30)
        psd = arma2psd(A=a, B=b, rho=rho, sides='centerdc', norm=True)
        pylab.plot(10 * pylab.log10(psd))
        pylab.ylim([-50,0])

    :reference: [Marple]_
    """
    R = CORRELATION(X, maxlags=lag, norm='unbiased')
    R0 = R[0]
    #C   Estimate the AR parameters (no error weighting is used).
    #C   Number of equation errors is M-Q .
    MPQ = lag - Q + P

    N = len(X)
    Y = np.zeros(N-P, dtype=complex)

    for K in range(0, MPQ):
        KPQ = K + Q - P+1
        if KPQ < 0:
            Y[K] = R[-KPQ].conjugate()
        if KPQ == 0:
            Y[K] = R0
        if KPQ > 0:
            Y[K] = R[KPQ]

    # The resize is very important for the normalissation.
    Y.resize(lag, refcheck=False)
    if P <= 4:
        res = arcovar_marple(Y.copy(), P)    #! Eq. (10.12)
        ar_params = res[0]
    else:
        res = arcovar(Y.copy(), P)    #! Eq. (10.12)
        ar_params = res[0]

    # the .copy is used to prevent a reference somewhere. this is a bug
    # to be tracked down.
    Y.resize(N-P, refcheck=False)

    #C   Filter the original time series
    for k in range(P, N):
        SUM = X[k]
        #SUM += sum([ar_params[j]*X[k-j-1] for j in range(0,P)])
        for j in range(0, P):
            SUM = SUM + ar_params[j] * X[k-j-1]   #! Eq. (10.17)
        Y[k-P] = SUM

    #  Estimate the MA parameters (a "long" AR of order at least 2*IQ
    #C   is suggested)
    #Y.resize(N-P)
    ma_params, rho = ma(Y, Q, 2*Q)     #! Eq. (10.3)

    return ar_params, ma_params, rho


class parma(ParametricSpectrum):
    """Class to create PSD using ARMA estimator.

    See :func:`arma_estimate` for description.

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import parma, marple_data
        p = parma(marple_data, 4, 4, 30, NFFT=4096)
        p.plot(sides='centerdc')


    """
    def __init__(self, data, P, Q, lag, NFFT=None, sampling=1.,
            scale_by_freq=False):
        """**Constructor:**

        For a detailed description of the parameters, see :func:`arma_estimate`.

        :param array data:     input data (list or numpy.array)
        :param int P:
        :param int Q:
        :param int lag:
        :param int NFFT: total length of the final data sets (padded with
            zero if needed; default is 4096)
        :param float sampling: sampling frequency of the input :attr:`data`.

        """
        super(parma, self).__init__(data, ma_order=Q, ar_order=P, lag=lag,
                                    NFFT=NFFT, sampling=sampling,
                                    scale_by_freq=scale_by_freq)
        self.lag = lag

    def __call__(self):
        ar_params, ma_params, rho = arma_estimate(self.data, self.ar_order,
                                         self.ma_order, self.lag)
        self.ma = ma_params
        self.ar = ar_params
        self.rho = rho
        psd = arma2psd(A=self.ar, B=self.ma, rho=self.rho,
                      T=self.sampling, NFFT=self.NFFT)
        #self.psd = psd
        if self.datatype == 'real':
            if self.NFFT % 2 == 0:
                newpsd  = psd[0:int(self.NFFT/2+1)] * 2
            else:
                newpsd  = psd[0:int((self.NFFT+1)/2)] * 2
            self.psd = newpsd
        else:
            self.psd = psd
        if self.scale_by_freq is True:
            self.scale()
        return self

    def _str_title(self):
        return "ARMA PSD estimate\n"

    def __str__(self):
        return super(parma, self).__str__()


class pma(ParametricSpectrum):
    """Class to create PSD using MA estimator.

    See :func:`ma` for description.

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import pma, marple_data
        p = pma(marple_data, 15, 30, NFFT=4096)
        p.plot(sides='centerdc')


    """
    def __init__(self, data, Q, M, NFFT=None, sampling=1.,
            scale_by_freq=False):
        """**Constructor:**

        For a detailed description of the parameters, see :func:`ma`.

        :param array data:     input data (list or numpy.array)
        :param int Q:          MA order
        :param int M:          AR model used to estimate the MA parameters
        :param int NFFT:       total length of the final data sets (padded with zero 
                               if needed; default is 4096)
        :param float sampling: sampling frequency of the input :attr:`data`.

        """
        super(pma, self).__init__(data, ma_order=Q, ar_order=M,
                                  NFFT=NFFT, sampling=sampling,
                                  scale_by_freq=scale_by_freq)

    def __call__(self):
        ma_params, rho = ma(self.data, self.ma_order, self.ar_order)
        self.ma = ma_params
        self.rho = rho
        psd = arma2psd(A=None, B=self.ma, rho=self.rho,
                      T=self.sampling, NFFT=self.NFFT)
        if self.datatype == 'real':
            if self.NFFT % 2 == 0:
                newpsd  = psd[0:int(self.NFFT/2+1)] * 2
            else:
                newpsd  = psd[0:int((self.NFFT+1)/2)] * 2
            self.psd = newpsd
        else:
            self.psd = psd
        if self.scale_by_freq is True:
            self.scale()
        self.modified = False

    def _str_title(self):
        return "MA (moving average) PSD estimate\n"

    def __str__(self):
        return super(pma, self).__str__()


def ma(X, Q, M):
    """Moving average estimator.

    This program provides an estimate of the moving average parameters
    and driving noise variance for a data sequence based on a
    long AR model and a least squares fit.

    :param array X: The input data array
    :param int Q: Desired MA model order (must be >0 and <M)
    :param int M: Order of "long" AR model (suggest at least 2*Q )

    :return:
        * MA    - Array of Q complex MA parameter estimates
        * RHO   - Real scalar of white noise variance estimate

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import arma2psd, ma, marple_data
        import pylab

        # Estimate 15 Ma parameters
        b, rho = ma(marple_data, 15, 30)
        # Create the PSD from those MA parameters
        psd = arma2psd(B=b, rho=rho, sides='centerdc')
        # and finally plot the PSD
        pylab.plot(pylab.linspace(-0.5, 0.5, 4096), 10 * pylab.log10(psd/max(psd)))
        pylab.axis([-0.5, 0.5, -30, 0])

    :reference: [Marple]_
    """
    if Q <= 0 or Q >= M:
        raise ValueError('Q(MA) must be in ]0,lag[')

    #C   Fit a high-order AR to the data
    a, rho, _c = yulewalker.aryule(X, M, 'biased')   #! Eq. (10.5)

    #add an element unity to the AR parameter array
    a = np.insert(a, 0, 1)

    #C   Find MA parameters from autocorrelations by Yule-Walker method
    ma_params, _p, _c = yulewalker.aryule(a, Q, 'biased')    #! Eq. (10.7)

    return ma_params, rho



"""
.. topic:: Toeplitz module

    These functions are not yet used by other functions, which explains the
    lack of documentation, test, examples. 
    
    Nevertheless, they can be used for production. 

    .. autosummary::

        HERMTOEP
        
    .. codeauthor:: Thomas Cokelaer, 2011
"""

import numpy

__all__ = [ "HERMTOEP"]


def TOEPLITZ(T0, TC, TR, Z):
    """solve the general toeplitz linear equations

    Solve TX=Z
    
    :param T0: zero lag value
    :param TC: r1 to rN 
    :param TR: r1 to rN

    returns X

    requires 3M^2+M operations instead of M^3 with gaussian elimination
    
    .. warning:: not used right now
    """
    assert len(TC)>0
    assert len(TC)==len(TR)
    M = len(TC)
    X = numpy.zeros(M+1,dtype=complex)
    A = numpy.zeros(M,dtype=complex)
    B = numpy.zeros(M,dtype=complex)
    P = T0
    if P == 0: raise ValueError("P must be different from zero")
    if P == 0: raise ValueError("P must be different from zero")
    X[0] = Z[0]/T0 
    for k in range(0, M):
        save1 = TC[k]
        save2 = TR[k]
        beta = X[0]*TC[k]
        if k == 0:
            temp1 = -save1 / P
            temp2 = -save2 / P
        else:
            for j in range(0, k):
                save1 = save1 + A[j] * TC[k-j-1]
                save2 = save2 + B[j] * TR[k-j-1]
                beta = beta + X[j+1] * TC[k-j-1]
            temp1 = -save1 / P
            temp2 = -save2/P
        P = P * (1. - (temp1*temp2))
        if P <= 0:
            raise ValueError("singular matrix")
        A[k] = temp1
        B[k] = temp2
        alpha = (Z[k+1]-beta)/P
        if k == 0: 
            X[k+1] = alpha
            for j in range(0,k+1):
                X[j] = X[j] + alpha * B[k-j]
            continue
        
        for j in range(0, k):
            kj = k-j-1
            save1 = A[j]
            A[j] = save1 + temp1 * B[kj] 
            B[kj] = B[kj] + temp2*save1
            
        X[k+1] = alpha
        for j in range(0,k+1):
            X[j] = X[j] + alpha*B[k-j]
    return X


def HERMTOEP(T0, T, Z):
    """solve Tx=Z by a variation of Levinson algorithm where T 
    is a complex hermitian toeplitz matrix

    :param T0: zero lag value
    :param T: r1 to rN
     
    :return: X
    
    used by eigen PSD method
    """
    assert len(T)>0
    M = len(T)
    X = numpy.zeros(M+1,dtype=complex)
    A = numpy.zeros(M,dtype=complex)
    P = T0
    if P == 0: raise ValueError("P must be different from zero")
    X[0] = Z[0]/T0 
    for k in range(0, M):
        save = T[k]
        beta = X[0]*T[k]
        if k == 0:
            temp = -save / P
        else:
            for j in range(0, k):
                save = save + A[j] * T[k-j-1]
                beta = beta + X[j+1] * T[k-j-1]
            temp = -save / P
        P = P * (1. - (temp.real**2+temp.imag**2))
        if P <= 0:
            raise ValueError("singular matrix")
        A[k] = temp
        alpha = (Z[k+1]-beta)/P

        if k == 0:
            #print 'skipping code for k=0' 
            X[k+1] = alpha
            for j in range(0,k+1):
                X[j] = X[j] + alpha * A[k-j].conjugate()
            continue
        khalf = (k+1)//2
        for j in range(0, khalf):
            kj = k-j-1
            save=A[j]
            A[j] = save+temp*A[kj].conjugate() 
            if j != kj:
                A[kj] = A[kj] + temp*save.conjugate()
        X[k+1] = alpha
        for j in range(0,k+1):
            X[j] = X[j] + alpha * A[k-j].conjugate()
    return X





"""
.. topic:: This module provides Periodograms (classics, daniell, bartlett)


    .. autosummary::

        Periodogram
        DaniellPeriodogram
        speriodogram
        WelchPeriodogram
        speriodogram

    .. codeauthor:: Thomas Cokelaer 2011

    :References: See [Marple]_

.. rubric:: Usage

You can compute a periodogram using :func:`speriodogram`::

    from spectrum import speriodogram, marple_data
    from pylab import plot
    p = speriodogram(marple_data)
    plot(p)

However, the output is not always easy to manipulate or plot, therefore
it is advised to use the class :class:`Periodogram` instead::

    from spectrum import Periodogram, marple_data
    p = Periodogram(marple_data)
    p.plot()

This class will take care of the plotting and internal state of
the computation. For instance, if you can change the output easily::

    p.plot(sides='twosided')

"""
import logging

from .window import Window
from .psd import Spectrum, FourierSpectrum
from numpy.fft import fft, rfft
import numpy as np


__all__ = ['pdaniell', 'speriodogram', 'Periodogram', 'WelchPeriodogram',
           'DaniellPeriodogram']


def speriodogram(x, NFFT=None, detrend=True, sampling=1.,
                   scale_by_freq=True, window='hamming', axis=0):
    """Simple periodogram, but matrices accepted.

    :param x: an array or matrix of data samples.
    :param NFFT: length of the data before FFT is computed (zero padding)
    :param bool detrend: detrend the data before co,puteing the FFT
    :param float sampling: sampling frequency of the input :attr:`data`.

    :param scale_by_freq:
    :param str window:

    :return: 2-sided PSD if complex data, 1-sided if real.

    if a matrix is provided (using numpy.matrix), then a periodogram
    is computed for each row. The returned matrix has the same shape as the input
    matrix.

    The mean of the input data is also removed from the data before computing
    the psd.

    .. plot::
        :width: 80%
        :include-source:

        from pylab import grid, semilogy
        from spectrum import data_cosine, speriodogram
        data = data_cosine(N=1024, A=0.1, sampling=1024, freq=200)
        semilogy(speriodogram(data, detrend=False, sampling=1024), marker='o')
        grid(True)


    .. plot::
        :width: 80%
        :include-source:

        import numpy
        from spectrum import speriodogram, data_cosine
        from pylab import figure, semilogy, figure ,imshow
        # create N data sets and make the frequency dependent on the time
        N = 100
        m = numpy.concatenate([data_cosine(N=1024, A=0.1, sampling=1024, freq=x) 
            for x in range(1, N)]);
        m.resize(N, 1024)
        res = speriodogram(m)
        figure(1)
        semilogy(res)
        figure(2)
        imshow(res.transpose(), aspect='auto')

    .. todo:: a proper spectrogram class/function that takes care of normalisation
    """
    x = np.array(x)
    # array with 1 dimension case
    if x.ndim == 1:
        axis = 0
        r = x.shape[0]
        w = Window(r, window)   #same size as input data
        w = w.data
    # matrix case
    elif x.ndim == 2:
        logging.debug('2D array. each row is a 1D array')
        [r, c] = x.shape
        w = np.array([Window(r, window).data for this in range(c)]).reshape(r,c) 

    if NFFT is None:
        NFFT = len(x)

    isreal = np.isrealobj(x)

    if detrend == True:
        m = np.mean(x, axis=axis)
    else:
        m = 0

    if isreal == True:
        if x.ndim == 2:
            res =  (abs (rfft (x*w - m, NFFT, axis=0))) ** 2. / r
        else:
            res =  (abs (rfft (x*w - m, NFFT, axis=-1))) ** 2. / r
    else:
        if x.ndim == 2:
            res =  (abs (fft (x*w - m, NFFT, axis=0))) ** 2. / r
        else:
            res =  (abs (fft (x*w - m, NFFT, axis=-1))) ** 2. / r

    if scale_by_freq is True:
        df = sampling / float(NFFT)
        res*= 2 * np.pi / df

    if x.ndim == 1:
        return res.transpose()
    else:
        return res


def WelchPeriodogram(data, NFFT=None,  sampling=1., **kargs):
    r"""Simple periodogram wrapper of numpy.psd function.

    :param A: the input data
    :param int NFFT: total length of the final data sets (padded 
        with zero if needed; default is 4096)
    :param str window:

    :Technical documentation:

    When we calculate the periodogram of a set of data we get an estimation
    of the spectral density. In fact as we use a Fourier transform and a
    truncated segments the spectrum is the convolution of the data with a
    rectangular window which Fourier transform is

    .. math::

        W(s)= \frac{1}{N^2} \left[ \frac{\sin(\pi s)}{\sin(\pi s/N)} \right]^2

    Thus oscillations and sidelobes appears around the main frequency. One aim of t he tapering is to reduced this effects. We multiply data by a window whose  sidelobes are much smaller than the main lobe. Classical window is hanning window.  But other windows are available. However we must take into account this energy and divide the spectrum by energy of taper used. Thus periodogram becomes :

    .. math::

        D_k \equiv \sum_{j=0}^{N-1}c_jw_j \; e^{2\pi ijk/N}  \qquad k=0,...,N-1

    .. math::

        P(0)=P(f_0)=\frac{1}{2\pi W_{ss}}\arrowvert{D_0}\arrowvert^2

    .. math::

        P(f_k)=\frac{1}{2\pi W_{ss}} \left[\arrowvert{D_k}\arrowvert^2+\arrowvert{D_{N-k}}\arrowvert^2\right]        \qquad k=0,1,...,     \left( \frac{1}{2}-1 \right)

    .. math::

        P(f_c)=P(f_{N/2})= \frac{1}{2\pi W_{ss}} \arrowvert{D_{N/2}}\arrowvert^2

    with

    .. math::

        {W_{ss}} \equiv N\sum_{j=0}^{N-1}w_j^2


    .. plot::
        :width: 80%
        :include-source:

        from spectrum import WelchPeriodogram, marple_data
        psd = WelchPeriodogram(marple_data, 256)

    """
    from pylab import psd
    spectrum = Spectrum(data, sampling=1.)

    P = psd(data, NFFT, Fs=sampling, **kargs)
    spectrum.psd = P[0]
    #spectrum.__Spectrum_sides = 'twosided'

    return P, spectrum


class Periodogram(FourierSpectrum):
    """The Periodogram class provides an interface to periodogram PSDs

    .. plot::
        :width: 80%
        :include-source:

        from spectrum import Periodogram, data_cosine
        data = data_cosine(N=1024, A=0.1, sampling=1024, freq=200)
        p = Periodogram(data, sampling=1024)
        p.plot(marker='o')


    """
    def __init__(self, data, sampling=1.,
                 window='hann', NFFT=None, scale_by_freq=False,
                 detrend=None):
        """**Periodogram Constructor**

        :param array data: input data (list or numpy.array)
        :param float sampling: sampling frequency of the input :attr:`data`.
        :param str window: a tapering window. See :class:`~spectrum.window.Window`.
        :param int NFFT: total length of the final data sets (padded with zero
            if needed; default is 4096)
        :param bool scale_by_freq:
        :param str detrend:

        """
        super(Periodogram, self).__init__(data,
                                          window=window,
                                          sampling=sampling,
                                          NFFT=NFFT,
                                          scale_by_freq=scale_by_freq,
                                          detrend=detrend)

    def __call__(self):
        psd = speriodogram(self.data, window=self.window, sampling=self.sampling,
                             NFFT=self.NFFT, scale_by_freq=self.scale_by_freq,
                             detrend=self.detrend)
        self.psd = psd
        if self.scale_by_freq is True:
            self.scale()
        return self

    def _str_title(self):
        return "Periodogram PSD estimate\n"

    def __str__(self):
        return super(Periodogram, self).__str__()


def DaniellPeriodogram(data, P, NFFT=None, detrend='mean', sampling=1.,
                       scale_by_freq=True, window='hamming'):
    r"""Return Daniell's periodogram.

    To reduce fast fluctuations of the spectrum one idea proposed by daniell
    is to average each value with points in its neighboorhood. It's like
    a low filter.

    .. math:: \hat{P}_D[f_i]= \frac{1}{2P+1} \sum_{n=i-P}^{i+P} \tilde{P}_{xx}[f_n]

    where P is the number of points to average.

    Daniell's periodogram is the convolution of the spectrum with a low filter:

    .. math:: \hat{P}_D(f)=   \hat{P}_{xx}(f)*H(f)

    Example::

        >>> DaniellPeriodogram(data, 8)

    if N/P is not integer, the final values of the original PSD are not used.

    using DaniellPeriodogram(data, 0) should give the original PSD.

    """
    psd = speriodogram(data, NFFT=NFFT, detrend=detrend, sampling=sampling,
                   scale_by_freq=scale_by_freq, window=window)

    if len(psd) % 2 == 1:
        datatype = 'real'
    else:
        datatype = 'complex'

    N = len(psd)
    _slice = 2 * P + 1
    if datatype == 'real': #must get odd value
        newN = np.ceil(psd.size/float(_slice))
        if newN % 2 == 0:
            newN = psd.size/_slice
    else:
        newN = np.ceil(psd.size/float(_slice))
        if newN % 2 == 1:
            newN = psd.size/_slice

    newpsd = np.zeros(int(newN)) # keep integer division
    for i in range(0, newpsd.size):
        count = 0 #needed to know the number of valid averaged values
        for n in range(i*_slice-P, i*_slice+P+1): #+1 to have P values on each sides
            if n > 0 and n<N: #needed to start the average
                count += 1
                newpsd[i] += psd[n]
        newpsd[i] /= float(count)

    #todo: check this
    if datatype == 'complex':
        freq = np.linspace(0, sampling, len(newpsd))
    else:
        df = 1. / sampling
        freq = np.linspace(0,sampling/2., len(newpsd))
    #psd.refreq(2*psd.size()/A.freq());
    #psd.retime(-1./psd.freq()+1./A.size());

    return newpsd, freq


class pdaniell(FourierSpectrum):
    """The pdaniell class provides an interface to DaniellPeriodogram

    ::

        from spectrum import data_cosine, pdaniell
        data = data_cosine(N=4096, sampling=4096)
        p = pdaniell(data, 8, NFFT=4096)
        p.plot()


    """
    def __init__(self, data, P, sampling=1.,
                 window='hann', NFFT=None, scale_by_freq=True,
                 detrend=None):
        """**pdaniell Constructor**

        :param array data:  input data (list or numpy.array)
        :param int P:       number of neighbours to average over.
        :param float sampling: sampling frequency of the input :attr:`data`.
        :param str window:  a tapering window. See :class:`~spectrum.window.Window`.
        :param int NFFT: total length of the final data sets (padded with 
            zero if needed; default is 4096)
        :param bool scale_by_freq:
        :param str detrend:

        """
        super(pdaniell, self).__init__(data,
                                       window=window,
                                       sampling=sampling,
                                       NFFT=NFFT,
                                       scale_by_freq=scale_by_freq,
                                       detrend=detrend)
        self.P = P
    def __call__(self):
        res = DaniellPeriodogram(self.data, self.P, window=self.window,
                  sampling=self.sampling, NFFT=self.NFFT,
                  scale_by_freq=self.scale_by_freq,
                  detrend=self.detrend)
        self.psd = res[0]
        return self

    def _str_title(self):
        return "Daniell Periodogram PSD estimate\n"

    def __str__(self):
        return super(pdaniell, self).__str__()








class SpectrumError(Exception):
    """A Base class error

    """
    #def __init__(self, value=None):
    #    self.value = value
    def __str__(self):
        return "Unknown Spectrum Error:"


class SpectrumModifiedError(SpectrumError):
    def __str__(self):
        msg = """Data has been modified (or NFFT, or sampling, or N) but
            the PSD has not be (re)-computed. Call it using () syntax (.e.g. p =
parma(...); p() )"""
        return msg


class SpectrumChoiceError(SpectrumError):
    def __init__(self, value, valid):
        #super(SpectrumChoiceError, self).__init__()
        self.value = value
        self.valid = valid
    def __str__(self):
        msg = """Spectrum error: invalid choice (%s). Possible values are %s.""" %\
            (self.value, self.valid)
        return msg


class SpectrumPSDError(SpectrumError):
    def __str__(self):
        msg = """PSD not computed. Populate the attribute psd."""
        return msg


class SpectrumARMAError(SpectrumError):
    def __str__(self):
        msg = """Either AR parameter and/or MA must be provided"""
        return msg


class SpectrumARError(SpectrumError):
    def __str__(self):
        msg = """AR order must be stricly positive"""
        return msg


class SpectrumMAError(SpectrumError):
    def __str__(self):
        msg = """MA order must be stricly positive"""
        return msg


class SpectrumOrder(SpectrumError):
    def __str__(self):
        msg = """AR or MA order must be stricly positive"""
        return msg


class SpectrumNFFT(SpectrumError):
    def __str__(self):
        msg = """NFFT must be stricly positive"""
        return msg


def is_positive_integer(order, name="wrong argument. "):
    if order < 0:
        raise SpectrumOrder
    if type(order) != int:
        raise TypeError("%s must be an integer" % name)
    return True



"""
.. topic:: Cholesky methods

    .. autosummary::

        CHOLESKY

    .. codeauthor:: Thomas Cokelaer, 2011
"""
import numpy


__all__ = ["CHOLESKY"]


def _numpy_cholesky(A, B):
    """Solve Ax=B using numpy cholesky solver

    A = LU

    in the case where A is square and Hermitian, A = L.L* where L* is
    transpoed and conjugate matrix

    Ly = b

    where

    Ux=y

    so x = U^{-1} y
    where U = L*
    and y = L^{-1} B
    """
    L = numpy.linalg.cholesky(A)
    # A=L*numpy.transpose(L).conjugate()
    # Ly = b
    y = numpy.linalg.solve(L,B)
    # Ux = y
    x = numpy.linalg.solve(L.transpose().conjugate(),y)
    return x, L

def _numpy_solver(A, B):
    """This function solve Ax=B directly without taking care of the input
    matrix properties.
    """
    x = numpy.linalg.solve(A, B)
    return x

def CHOLESKY(A, B, method='scipy'):
    """Solve linear system `AX=B` using CHOLESKY method.

    :param A: an input Hermitian matrix
    :param B: an array
    :param str method: a choice of method in [numpy, scipy, numpy_solver]

        * `numpy_solver` relies entirely on numpy.solver (no cholesky decomposition)
        * `numpy` relies on the numpy.linalg.cholesky for the decomposition and
          numpy.linalg.solve for the inversion.
        * `scipy` uses scipy.linalg.cholesky for the decomposition and
          scipy.linalg.cho_solve for the inversion.

    .. rubric:: Description

    When a matrix is square and Hermitian (symmetric with lower part being
    the complex conjugate of the upper one), then the usual triangular
    factorization takes on the special form:

    .. math:: A = R R^H

    where :math:`R` is a lower triangular matrix with nonzero real principal
    diagonal element. The input matrix can be made of complex data. Then, the
    inversion to find :math:`x` is made as follows:

    .. math::  Ry = B

    and

    .. math::   Rx = y

    .. doctest::

        >>> import numpy
        >>> from spectrum import CHOLESKY
        >>> A = numpy.array([[ 2.0+0.j ,  0.5-0.5j, -0.2+0.1j],
        ...    [ 0.5+0.5j,  1.0+0.j ,  0.3-0.2j],
        ...    [-0.2-0.1j,  0.3+0.2j,  0.5+0.j ]])
        >>> B = numpy.array([ 1.0+3.j ,  2.0-1.j ,  0.5+0.8j])
        >>> CHOLESKY(A, B)
        array([ 0.95945946+5.25675676j,  4.41891892-7.04054054j,
               -5.13513514+6.35135135j])

    """
    if method == 'numpy_solver':
        X = _numpy_solver(A,B)
        return X
    elif method == 'numpy':
        X, _L = _numpy_cholesky(A, B)
        return X
    elif method == 'scipy':
        import scipy.linalg
        L = scipy.linalg.cholesky(A)
        X = scipy.linalg.cho_solve((L, False), B)
    else:
        raise ValueError('method must be numpy_solver, numpy_cholesky or cholesky_inplace')
    return X



"""
.. topic:: Correlation module


    Provides two correlation functions. :func:`CORRELATION` is slower than
    :func:`xcorr`. However, the output is as expected by some other functions.
    Ultimately, it should be replaced by :func:`xcorr`.

    For real data, the behaviour of the 2 functions is identical. However, for
    complex data, xcorr returns a 2-sides correlation.


    .. autosummary::

        ~spectrum.correlation.CORRELATION
        ~spectrum.correlation.xcorr

    .. codeauthor: Thomas Cokelaer, 2011



"""
import numpy as np


__all__ = ['CORRELATION', 'xcorr']


def pylab_rms_flat(a):
    """
    Return the root mean square of all the elements of *a*, flattened out.
    (Copied 1:1 from matplotlib.mlab.)
    """
    return np.sqrt(np.mean(np.absolute(a) ** 2))


def CORRELATION(x, y=None, maxlags=None, norm='unbiased'):
    r"""Correlation function

    This function should give the same results as :func:`xcorr` but it
    returns the positive lags only. Moreover the algorithm does not use
    FFT as compared to other algorithms.

    :param array x: first data array of length N
    :param array y: second data array of length N. If not specified, computes the
        autocorrelation.
    :param int maxlags: compute cross correlation between [0:maxlags]
        when maxlags is not specified, the range of lags is [0:maxlags].
    :param str norm: normalisation in ['biased', 'unbiased', None, 'coeff']

        * *biased*   correlation=raw/N,
        * *unbiased* correlation=raw/(N-`|lag|`)
        * *coeff*    correlation=raw/(rms(x).rms(y))/N
        * None       correlation=raw

    :return:
        * a numpy.array correlation sequence,  r[1,N]
        * a float for the zero-lag correlation,  r[0]

    The *unbiased* correlation has the form:

    .. math::

        \hat{r}_{xx} = \frac{1}{N-m}T \sum_{n=0}^{N-m-1} x[n+m]x^*[n] T

    The *biased* correlation differs by the front factor only:

    .. math::

        \check{r}_{xx} = \frac{1}{N}T \sum_{n=0}^{N-m-1} x[n+m]x^*[n] T

    with :math:`0\leq m\leq N-1`.

    .. doctest::

        >>> from spectrum import CORRELATION
        >>> x = [1,2,3,4,5]
        >>> res = CORRELATION(x,x, maxlags=0, norm='biased')
        >>> res[0]
        11.0

    .. note:: this function should be replaced by :func:`xcorr`.

    .. seealso:: :func:`xcorr`
    """
    assert norm in ['unbiased','biased', 'coeff', None]
    #transform lag into list if it is an integer
    x = np.array(x)
    if y is None:
        y = x
    else:
        y = np.array(y)

    # N is the max of x and y
    N = max(len(x), len(y))
    if len(x) < N:
        x = y.copy()
        x.resize(N, refcheck=False)
    if len(y) < N:
        y = y.copy()
        y.resize(N, refcheck=False)

    #default lag is N-1
    if maxlags is None:
        maxlags = N - 1
    assert maxlags < N, 'lag must be less than len(x)'

    realdata = np.isrealobj(x) and np.isrealobj(y)
    #create an autocorrelation array with same length as lag
    if realdata == True:
        r = np.zeros(maxlags, dtype=float)
    else:
        r = np.zeros(maxlags, dtype=complex)

    if norm == 'coeff':
        rmsx = pylab_rms_flat(x)
        rmsy = pylab_rms_flat(y)

    for k in range(0, maxlags+1):
        nk = N - k - 1

        if realdata == True:
            sum = 0
            for j in range(0, nk+1):
                sum = sum + x[j+k] * y[j]
        else:
            sum = 0. + 0j
            for j in range(0, nk+1):
                sum = sum + x[j+k] * y[j].conjugate()
        if k == 0:
            if norm in ['biased', 'unbiased']:
                r0 = sum/float(N)
            elif norm is None:
                r0 = sum
            else:
                r0 =  1.
        else:
            if norm == 'unbiased':
                r[k-1] = sum / float(N-k)
            elif norm == 'biased':
                r[k-1] = sum / float(N)
            elif norm is None:
                r[k-1] = sum
            elif norm == 'coeff':
                r[k-1] =  sum/(rmsx*rmsy)/float(N)

    r = np.insert(r, 0, r0)
    return r


def xcorr(x, y=None, maxlags=None, norm='biased'):
    """Cross-correlation using numpy.correlate

    Estimates the cross-correlation (and autocorrelation) sequence of a random
    process of length N. By default, there is no normalisation and the output
    sequence of the cross-correlation has a length 2*N+1.

    :param array x: first data array of length N
    :param array y: second data array of length N. If not specified, computes the
        autocorrelation.
    :param int maxlags: compute cross correlation between [-maxlags:maxlags]
        when maxlags is not specified, the range of lags is [-N+1:N-1].
    :param str option: normalisation in ['biased', 'unbiased', None, 'coeff']

    The true cross-correlation sequence is

    .. math:: r_{xy}[m] = E(x[n+m].y^*[n]) = E(x[n].y^*[n-m])

    However, in practice, only a finite segment of one realization of the
    infinite-length random process is available.

    The correlation is estimated using numpy.correlate(x,y,'full').
    Normalisation is handled by this function using the following cases:

        * 'biased': Biased estimate of the cross-correlation function
        * 'unbiased': Unbiased estimate of the cross-correlation function
        * 'coeff': Normalizes the sequence so the autocorrelations at zero
           lag is 1.0.

    :return:
        * a numpy.array containing the cross-correlation sequence (length 2*N-1)
        * lags vector

    .. note:: If x and y are not the same length, the shorter vector is
        zero-padded to the length of the longer vector.

    .. rubric:: Examples

    .. doctest::

        >>> from spectrum import xcorr
        >>> x = [1,2,3,4,5]
        >>> c, l = xcorr(x,x, maxlags=0, norm='biased')
        >>> c
        array([ 11.])

    .. seealso:: :func:`CORRELATION`.
    """
    N = len(x)
    if y is None:
        y = x
    assert len(x) == len(y), 'x and y must have the same length. Add zeros if needed'

    if maxlags is None:
        maxlags = N-1
        lags = np.arange(0, 2*N-1)
    else:
        assert maxlags <= N, 'maxlags must be less than data length'
        lags = np.arange(N-maxlags-1, N+maxlags)

    res = np.correlate(x, y, mode='full')

    if norm == 'biased':
        Nf = float(N)
        res = res[lags] / float(N)    # do not use /= !!
    elif norm == 'unbiased':
        res = res[lags] / (float(N)-abs(np.arange(-N+1, N)))[lags]
    elif norm == 'coeff':
        Nf = float(N)
        rms = pylab_rms_flat(x) * pylab_rms_flat(y)
        res = res[lags] / rms / Nf
    else:
        res = res[lags]

    lags = np.arange(-maxlags, maxlags+1)
    return res, lags



"""
.. topic:: Linear algebra tools

    .. autosummary::

        csvd
        corrmtx
        pascal

    .. codeauthor:: Thomas Cokelaer 2011


"""
import numpy

import spectrum.errors as errors

__all__ = ["pascal", "csvd", "corrmtx"]


def pascal(n):
    """Return Pascal matrix

    :param int n: size of the matrix

    .. doctest::

        >>> from spectrum import pascal
        >>> pascal(6)
        array([[   1.,    1.,    1.,    1.,    1.,    1.],
               [   1.,    2.,    3.,    4.,    5.,    6.],
               [   1.,    3.,    6.,   10.,   15.,   21.],
               [   1.,    4.,   10.,   20.,   35.,   56.],
               [   1.,    5.,   15.,   35.,   70.,  126.],
               [   1.,    6.,   21.,   56.,  126.,  252.]])

    .. todo:: use the symmetric property to improve computational time if needed
    """
    errors.is_positive_integer(n)
    result = numpy.zeros((n, n))

    #fill the first row and column
    for i in range(0, n):
        result[i, 0] = 1
        result[0, i] = 1
    if n > 1:
        for i in range(1, n):
            for j in range(1, n):
                result[i, j] = result[i-1, j] + result[i, j-1]
    return result



def corrmtx(x_input, m, method='autocorrelation'):
    r"""Correlation matrix

    This function is used by PSD estimator functions. It generates
    the correlation matrix from a correlation data set and a maximum lag.

    :param array x: autocorrelation samples (1D)
    :param int m: the maximum lag

    Depending on the choice of the method, the correlation matrix has different
    sizes, but the number of rows is always m+1.

    Method can be :

    * 'autocorrelation': (default) X is the (n+m)-by-(m+1) rectangular Toeplitz
      matrix derived using prewindowed and postwindowed data.
    * 'prewindowed': X is the n-by-(m+1) rectangular Toeplitz matrix derived
      using prewindowed data only.
    * 'postwindowed': X is the n-by-(m+1) rectangular Toeplitz matrix that
      derived using postwindowed data only.
    * 'covariance': X is the (n-m)-by-(m+1) rectangular Toeplitz matrix
      derived using nonwindowed data.
    * 'modified': X is the 2(n-m)-by-(m+1) modified rectangular Toeplitz
      matrix that generates an autocorrelation estimate for the length n data
      vector x, derived using forward and backward prediction error estimates.


    :return:
        * the autocorrelation matrix
        * R, the (m+1)-by-(m+1) autocorrelation matrix estimate ``R= X'*X``.

    .. rubric:: Algorithm details:

    The **autocorrelation** matrix is a :math:`(N+p) \times (p+1)` rectangular Toeplitz
    data matrix:

    .. math:: X_p = \begin{pmatrix}L_p\\T_p\\Up\end{pmatrix}

    where the lower triangular :math:`p \times (p+1)` matrix :math:`L_p` is

    .. math:: L_p =
        \begin{pmatrix}
        x[1]     &  \cdots     & 0     & 0        \\
        \vdots    &  \ddots     & \vdots & \vdots    \\
        x[p]     &  \cdots     & x[1]  & 0
        \end{pmatrix}

    where the rectangular :math:`(N-p) \times (p+1)` matrix :math:`T_p` is

    .. math:: T_p =
        \begin{pmatrix}
        x[p+1]     &  \cdots    & x[1]        \\
        x[p+2]     &  \cdots    & x[2]        \\
        \vdots    &  \ddots     & \vdots      \\
        x[N-1]     &  \cdots    & x[N-p-1]    \\
        x[N]     &  \cdots      & x[N-p]
        \end{pmatrix}

    and where the upper triangular :math:`p \times (p+1)` matrix :math:`U_p` is

    .. math:: U_p =
        \begin{pmatrix}
        0         &  x[N]      & \cdots     & x[N-p+1]        \\
        \vdots    &  \vdots    & \ddots & \vdots    \\
        0         &  0         & \cdots  & x[N]
        \end{pmatrix}

    From this definition, the prewindowed matrix is

    .. math:: X_p = \begin{pmatrix}L_p\\T_p\end{pmatrix}

    the postwindowed matrix is

    .. math:: X_p = \begin{pmatrix}T_p\\U_p\end{pmatrix}

    the covariance matrix is:

    .. math:: X_p = \begin{pmatrix}T_p\end{pmatrix}

    and the modified covariance matrix is:

    .. math:: X_p = \begin{pmatrix}T_p\\T_p^*\end{pmatrix}


    """

    valid_methods = ['autocorrelation', 'prewindowed', 'postwindowed',
                      'covariance', 'modified']
    if method not in valid_methods:
        raise ValueError("Invalid method. Try one of %s" % valid_methods)

    from scipy.linalg import toeplitz

    # create the relevant matrices that will be useful to create
    # the correlation matrices
    N = len(x_input)

    # FIXME:do we need a copy ?
    if isinstance(x_input, list):
        x = numpy.array(x_input)
    else:
        x = x_input.copy()


    if x.dtype == complex:
        complex_type = True
    else:
        complex_type = False

    # Compute the Lp, Up and Tp matrices according to the requested method
    if method in ['autocorrelation', 'prewindowed']:
        Lp = toeplitz(x[0:m], [0]*(m+1))
    Tp = toeplitz(x[m:N], x[m::-1])
    if method in ['autocorrelation', 'postwindowed']:
        Up = toeplitz([0]*(m+1), numpy.insert(x[N:N-m-1:-1],0,0))



    # Create the output matrix

    if method == 'autocorrelation':
        if complex_type == True:
            C = numpy.zeros((N+m, m+1), dtype=complex)
        else:
            C = numpy.zeros((N+m, m+1))
        for i in range(0, m):
            C[i] = Lp[i]
        for i in range(m, N):
            C[i] = Tp[i-m]
        for i in range(N, N+m):
            C[i] = Up[i-N]
    elif method == 'prewindowed':
        if complex_type == True:
            C = numpy.zeros((N, m+1), dtype=complex)
        else:
            C = numpy.zeros((N, m+1))

        for i in range(0, m):
            C[i] = Lp[i]
        for i in range(m, N):
            C[i] = Tp[i-m]
    elif method == 'postwindowed':
        if complex_type == True:
            C = numpy.zeros((N, m+1), dtype=complex)
        else:
            C = numpy.zeros((N, m+1))
        for i in range(0, N-m):
            C[i] = Tp[i]
        for i in range(N-m, N):
            C[i] = Up[i-N+m]
    elif method == 'covariance':
        return Tp
    elif method == 'modified':
        if complex_type == True:
            C = numpy.zeros((2*(N-m), m+1), dtype=complex)
        else:
            C = numpy.zeros((2*(N-m), m+1))
        for i in range(0, N-m):
            C[i] = Tp[i]
        Tp = numpy.fliplr(Tp.conj())
        for i in range(N-m, 2*(N-m)):
            C[i] = Tp[i-N+m]



    return C
    #autocorrelation n+m by m+1 matrix (Lp, Tp, Up)

    #Rx = toeplitz(r)



def csvd(A):
    """SVD decomposition using numpy.linalg.svd

    :param A: a M by N matrix

    :return:
        * U, a M by M matrix
        * S the N eigen values
        * V a N by N matrix

    See :func:`numpy.linalg.svd` for a detailed documentation.

    Should return the same as in [Marple]_ , CSVD routine.

    ::

        U, S, V = numpy.linalg.svd(A)
        U, S, V = cvsd(A)

    """
    U, S, V = numpy.linalg.svd(A)
    return U, S, V



from __future__ import absolute_import

try:
    import pkg_resources
    version = pkg_resources.require("spectrum")[0].version
    __version__ = version
except:
    __version__ = "?"



import logging
def spectrum_set_level(level):
    assert level in ['DEBUG', 'INFO', 'CRITICAL', 'ERROR', 'WARNING']
    logging.getLogger().setLevel(level)


#: default number of samples used to compute FFT
default_NFFT = 4096

from .arma import *
from .burg import *
from .cholesky import *
from .correlation import *
from .correlog import *
from .covar import *
from .criteria import *
from .datasets import *
from .eigen import *
from .eigenfre import *
from .io import *
from .levinson import *
from .linear_prediction import *
from .linalg import *
#from lms import *
from .lpc import *
from .minvar import *
from .modcovar import *
from .mtm import *
from .periodogram import *
from .psd import *
from .spectrogram import *
from .tools import *
from .toeplitz import *
from .transfer import *
from .window import *
from .waveform import *
from .yulewalker import *

